# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"vanilla_launcher contains some functions for interacting with the Vanilla Minecraft Launcher"
from ._internal_types.vanilla_launcher_types import (
    VanillaLauncherProfilesJson,
    VanillaLauncherProfilesJsonProfile,
)
from ._types import VanillaLauncherProfile, MinecraftOptions
from .exceptions import InvalidVanillaLauncherProfile
from .utils import get_latest_version
import datetime
import aiofiles
import json
import uuid
import os

__all__ = [
    "load_vanilla_launcher_profiles",
    "vanilla_launcher_profile_to_minecraft_options",
    "get_vanilla_launcher_profile_version",
    "add_vanilla_launcher_profile",
    "do_vanilla_launcher_profiles_exists",
]


async def _is_vanilla_launcher_profile_valid(
    vanilla_profile: VanillaLauncherProfile,
) -> bool:
    "Checks if the given profile is valid"
    # 名稱必須是字符串
    if not isinstance(vanilla_profile.get("name"), str):
        return False

    # 版本類型檢查
    if vanilla_profile.get("versionType") not in (
        "latest-release",
        "latest-snapshot",
        "custom",
    ):
        return False

    # custom 類型必須有 version
    if (
        vanilla_profile["versionType"] == "custom"
        and vanilla_profile.get("version") is None
    ):
        return False

    # 檢查 gameDirectory 類型
    if vanilla_profile.get("gameDirectory") is not None and not isinstance(
        vanilla_profile.get("gameDirectory"), str
    ):
        return False

    # 檢查 javaExecutable 類型
    if vanilla_profile.get("javaExecutable") is not None and not isinstance(
        vanilla_profile.get("javaExecutable"), str
    ):
        return False

    # 檢查 javaArguments
    java_arguments = vanilla_profile.get("javaArguments")
    if java_arguments is not None:
        if not all(isinstance(arg, str) for arg in java_arguments):
            return False

    # 檢查自定義分辨率
    custom_resolution = vanilla_profile.get("customResolution")
    if custom_resolution is not None:
        try:
            if len(custom_resolution) != 2:
                return False
            if not isinstance(custom_resolution["height"], int):
                return False
            if not isinstance(custom_resolution["width"], int):
                return False
        except Exception:
            return False

    return True


async def _read_launcher_profiles_json(
    minecraft_directory: str | os.PathLike,
) -> VanillaLauncherProfilesJson:
    """讀取 launcher_profiles.json 文件"""
    async with aiofiles.open(
        os.path.join(minecraft_directory, "launcher_profiles.json"),
        "r",
        encoding="utf-8",
    ) as f:
        data_text = await f.read()
        return json.loads(data_text)


async def _write_launcher_profiles_json(
    minecraft_directory: str | os.PathLike,
    data: VanillaLauncherProfilesJson,
) -> None:
    """寫入 launcher_profiles.json 文件"""
    async with aiofiles.open(
        os.path.join(minecraft_directory, "launcher_profiles.json"),
        "w",
        encoding="utf-8",
    ) as f:
        await f.write(json.dumps(data, ensure_ascii=False, indent=4))


async def load_vanilla_launcher_profiles(
    minecraft_directory: str | os.PathLike,
) -> list[VanillaLauncherProfile]:
    """
    Loads the profiles of the Vanilla Launcher from the given Minecraft directory

    :param minecraft_directory: The Minecraft directory
    :return: A List with the Profiles
    """
    data = await _read_launcher_profiles_json(minecraft_directory)

    profile_list: list[VanillaLauncherProfile] = []
    for value in data["profiles"].values():
        vanilla_profile: VanillaLauncherProfile = {}

        # 根據類型設置名稱
        vanilla_profile["name"] = (
            "Latest release"
            if value["type"] == "latest-release"
            else (
                "Latest snapshot"
                if value["type"] == "latest-snapshot"
                else value["name"]
            )
        )

        # 設置版本類型和版本
        version_id = value["lastVersionId"]
        if version_id == "latest-release":
            vanilla_profile["versionType"] = "latest-release"
            vanilla_profile["version"] = None
        elif version_id == "latest-snapshot":
            vanilla_profile["versionType"] = "latest-snapshot"
            vanilla_profile["version"] = None
        else:
            vanilla_profile["versionType"] = "custom"
            vanilla_profile["version"] = version_id

        # 設置額外屬性
        vanilla_profile["gameDirectory"] = value.get("gameDir")
        vanilla_profile["javaExecutable"] = value.get("javaDir")

        # 處理 Java 參數
        vanilla_profile["javaArguments"] = (
            value["javaArgs"].split(" ") if "javaArgs" in value else None
        )

        # 處理分辨率
        if "resolution" in value:
            vanilla_profile["customResolution"] = {
                "height": value["resolution"]["height"],
                "width": value["resolution"]["width"],
            }
        else:
            vanilla_profile["customResolution"] = None

        profile_list.append(vanilla_profile)

    return profile_list


async def vanilla_launcher_profile_to_minecraft_options(
    vanilla_profile: VanillaLauncherProfile,
) -> MinecraftOptions:
    """
    Converts a VanillaLauncherProfile into a Options dict, that can be used by :func:`~launcher_corecommand.get_minecraft_command`.
    You still need to add the Login Data to the Options before you can use it.

    :param vanilla_profile: The profile as returned by :func:`load_vanilla_launcher_profiles`
    :raises InvalidVanillaLauncherProfile: The given Profile is invalid
    :return: The Options Dict
    """
    if not await _is_vanilla_launcher_profile_valid(vanilla_profile):
        raise InvalidVanillaLauncherProfile(vanilla_profile)

    options: MinecraftOptions = {}

    # 只在值非 None 時添加到 options 字典中
    if (game_directory := vanilla_profile.get("gameDirectory")) is not None:
        options["gameDirectory"] = game_directory

    if (java_executable := vanilla_profile.get("javaExecutable")) is not None:
        options["executablePath"] = java_executable

    if (java_arguments := vanilla_profile.get("javaArguments")) is not None:
        options["jvmArguments"] = java_arguments

    if (custom_resolution := vanilla_profile.get("customResolution")) is not None:
        options["customResolution"] = True
        options["resolutionWidth"] = str(custom_resolution["width"])
        options["resolutionHeight"] = str(custom_resolution["height"])

    return options


async def get_vanilla_launcher_profile_version(
    vanilla_profile: VanillaLauncherProfile,
) -> str:
    """
    Returns the Minecraft version of the VanillaProfile. Handles ``latest-release`` and ``latest-snapshot``.

    :param vanilla_profile: The Profile
    :type vanilla_profile: VanillaLauncherProfile
    :raises InvalidVanillaLauncherProfile: The given Profile is invalid
    :return: The Minecraft version
    """
    if not await _is_vanilla_launcher_profile_valid(vanilla_profile):
        raise InvalidVanillaLauncherProfile(vanilla_profile)

    version_type = vanilla_profile["versionType"]

    if version_type == "latest-release":
        latest_version = await get_latest_version()
        return latest_version["release"]
    elif version_type == "latest-snapshot":
        latest_version = await get_latest_version()
        return latest_version["snapshot"]
    else:  # custom
        return vanilla_profile["version"]  # type: ignore


async def add_vanilla_launcher_profile(
    minecraft_directory: str | os.PathLike, vanilla_profile: VanillaLauncherProfile
) -> None:
    """
    Adds a new Profile to the Vanilla Launcher

    :param minecraft_directory: The Minecraft directory
    :param vanilla_profile: The new Profile
    :raises InvalidVanillaLauncherProfile: The given Profile is invalid
    """
    if not await _is_vanilla_launcher_profile_valid(vanilla_profile):
        raise InvalidVanillaLauncherProfile(vanilla_profile)

    data = await _read_launcher_profiles_json(minecraft_directory)

    new_profile: VanillaLauncherProfilesJsonProfile = {
        "name": vanilla_profile["name"],
        "created": datetime.datetime.now().isoformat(),
        "lastUsed": datetime.datetime.now().isoformat(),
        "type": "custom",
    }

    # 根據版本類型設置 lastVersionId
    version_type = vanilla_profile["versionType"]
    if version_type in ("latest-release", "latest-snapshot"):
        new_profile["lastVersionId"] = version_type
    else:  # custom
        new_profile["lastVersionId"] = vanilla_profile["version"]  # type: ignore

    # 設置可選參數
    if (game_directory := vanilla_profile.get("gameDirectory")) is not None:
        new_profile["gameDir"] = game_directory

    if (java_executable := vanilla_profile.get("javaExecutable")) is not None:
        new_profile["javaDir"] = java_executable

    if (java_arguments := vanilla_profile.get("javaArguments")) is not None:
        new_profile["javaArgs"] = " ".join(java_arguments)

    if (custom_resolution := vanilla_profile.get("customResolution")) is not None:
        new_profile["resolution"] = {
            "height": custom_resolution["height"],
            "width": custom_resolution["width"],
        }

    # 生成唯一的 key
    key = str(uuid.uuid4())
    while key in data["profiles"]:
        key = str(uuid.uuid4())

    data["profiles"][key] = new_profile

    await _write_launcher_profiles_json(minecraft_directory, data)


def do_vanilla_launcher_profiles_exists(
    minecraft_directory: str | os.PathLike,
) -> bool:
    """
    Checks if profiles from the vanilla launcher can be found

    :param minecraft_directory: The Minecraft directory
    :return: If profiles exists
    """
    return os.path.isfile(os.path.join(minecraft_directory, "launcher_profiles.json"))



import logging


def setup_logger(
    name: str = None,
    level: int = logging.INFO,
    filename: str = None,
    enable_console: bool = False,
) -> logging.Logger:
    """
    設定並返回一個 logger。

    :param name: Logger 的名稱，默認為 None（根 logger）。
    :param level: 日誌級別，默認為 logging.INFO。
    :param filename: 如果提供，日誌將寫入該文件。
    :param enable_console: 如果為 True，日誌將輸出到控制台。
    :return: 配置好的 logger。
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # 添加控制台處理器
    if enable_console:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    # 添加文件處理器
    if filename:
        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
from typing import TypedDict
import requests
import datetime


class RequestsResponseCache(TypedDict):
    response: requests.models.Response
    datetime: datetime.datetime


class MavenMetadata(TypedDict):
    release: str
    latest: str
    versions: list[str]



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"This module contains some helper functions. It should nt be used outside minecraft_launcher_lib"
from .exceptions import FileOutsideMinecraftDirectory, InvalidChecksum, VersionNotFound
from ._internal_types.shared_types import ClientJson, ClientJsonRule, ClientJsonLibrary
from ._internal_types.helper_types import RequestsResponseCache, MavenMetadata
from ._types import MinecraftOptions, CallbackDict
from typing import Literal, Any
import subprocess
import datetime
import aiohttp
import aiofiles
import platform
import hashlib
import zipfile
import lzma
import json
import sys
import re
import os

if os.name == "nt":
    info = subprocess.STARTUPINFO()  # type: ignore
    info.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # type: ignore
    info.wShowWindow = subprocess.SW_HIDE  # type: ignore
    SUBPROCESS_STARTUP_INFO: subprocess.STARTUPINFO | None = info  # type: ignore
else:
    SUBPROCESS_STARTUP_INFO = None


def empty(arg: Any) -> None:
    """
    This function is just a placeholder
    """
    pass


def check_path_inside_minecraft_directory(
    minecraft_directory: str | os.PathLike, path: str | os.PathLike
) -> None:
    """
    Raises a FileOutsideMinecraftDirectory if the Path is not in the given Directory
    """
    if not os.path.abspath(path).startswith(os.path.abspath(minecraft_directory)):
        raise FileOutsideMinecraftDirectory(
            os.path.abspath(path), os.path.abspath(minecraft_directory)
        )


async def download_file(
    url: str,
    path: str,
    callback: CallbackDict = {},
    sha1: str | None = None,
    lzma_compressed: bool | None = False,
    session: aiohttp.ClientSession | None = None,
    minecraft_directory: str | os.PathLike | None = None,
    overwrite: bool | None = False,
) -> bool:
    """
    Downloads a file into the given path. Check sha1 if given.
    """
    # Check if the Path is outside the given Minecraft Directory
    if minecraft_directory is not None:
        check_path_inside_minecraft_directory(minecraft_directory, path)

    if os.path.isfile(path) and not overwrite:
        if sha1 is None:
            return False
        elif await get_sha1_hash(path) == sha1:
            return False

    try:
        os.makedirs(os.path.dirname(path), exist_ok=True)
    except Exception:
        pass

    callback.get("setStatus", empty)("Download " + os.path.basename(path))

    close_session = False
    if session is None:
        session = aiohttp.ClientSession()
        close_session = True

    try:
        headers = {"user-agent": get_user_agent()}
        async with session.get(url, headers=headers) as r:
            if r.status != 200:
                return False

            content = await r.read()

            async with aiofiles.open(path, "wb") as f:
                if lzma_compressed:
                    await f.write(lzma.decompress(content))
                else:
                    await f.write(content)
    finally:
        if close_session:
            await session.close()

    if sha1 is not None:
        checksum = await get_sha1_hash(path)
        if checksum != sha1:
            raise InvalidChecksum(url, path, sha1, checksum)

    return True


def parse_single_rule(rule: ClientJsonRule, options: MinecraftOptions) -> bool:
    """
    Parse a single rule from the versions.json
    """
    if rule["action"] == "allow":
        returnvalue = False
    elif rule["action"] == "disallow":
        returnvalue = True

    for os_key, os_value in rule.get("os", {}).items():
        if os_key == "name":
            if os_value == "windows" and platform.system() != "Windows":
                return returnvalue
            elif os_value == "osx" and platform.system() != "Darwin":
                return returnvalue
            elif os_value == "linux" and platform.system() != "Linux":
                return returnvalue
        elif os_key == "arch":
            if os_value == "x86" and platform.architecture()[0] != "32bit":
                return returnvalue
        elif os_key == "version":
            if not re.match(os_value, get_os_version()):
                return returnvalue

    for features_key in rule.get("features", {}).keys():
        if features_key == "has_custom_resolution" and not options.get(
            "customResolution", False
        ):
            return returnvalue
        elif features_key == "is_demo_user" and not options.get("demo", False):
            return returnvalue
        elif (
            features_key == "has_quick_plays_support"
            and options.get("quickPlayPath") is None
        ):
            return returnvalue
        elif (
            features_key == "is_quick_play_singleplayer"
            and options.get("quickPlaySingleplayer") is None
        ):
            return returnvalue
        elif (
            features_key == "is_quick_play_multiplayer"
            and options.get("quickPlayMultiplayer") is None
        ):
            return returnvalue
        elif (
            features_key == "is_quick_play_realms"
            and options.get("quickPlayRealms") is None
        ):
            return returnvalue

    return not returnvalue


def parse_rule_list(rules: list[ClientJsonRule], options: MinecraftOptions) -> bool:
    """
    Parse a list of rules
    """
    for i in rules:
        if not parse_single_rule(i, options):
            return False

    return True


def _get_lib_name_without_version(lib: ClientJsonLibrary) -> str:
    """
    Returns the library name but without the version part
    e.g. org.ow2.asm:asm:9.7.1 -> org.ow2.asm:asm
    """
    return ":".join(lib["name"].split(":")[:-1])


async def inherit_json(
    original_data: ClientJson, path: str | os.PathLike
) -> ClientJson:
    """
    Implement the inheritsFrom function
    See https://github.com/tomsik68/mclauncher-api/wiki/Version-Inheritance-&-Forge
    """
    inherit_version = original_data["inheritsFrom"]

    file_path = os.path.join(
        path, "versions", inherit_version, inherit_version + ".json"
    )
    async with aiofiles.open(file_path, "r") as f:
        new_data: ClientJson = json.loads(await f.read())

    # Inheriting the libs is a bit special
    # If the lib is already present in the client.json in a different, it can't be inherited
    # So first we need a dict which contains all libs that are already present
    original_libs: dict[str, bool] = {}
    for current_lib in original_data.get("libraries", []):
        lib_name = _get_lib_name_without_version(current_lib)
        original_libs[lib_name] = True

    # Now we can attach all libs from the inherited version that are not already existing
    lib_list = original_data.get("libraries", [])
    for current_lib in new_data["libraries"]:
        lib_name = _get_lib_name_without_version(current_lib)
        if lib_name not in original_libs:
            lib_list.append(current_lib)

    new_data["libraries"] = lib_list

    for key, value in original_data.items():
        if key == "libraries":
            # We already had inherited the libs
            continue

        if isinstance(value, list) and isinstance(new_data.get(key, None), list):
            new_data[key] = value + new_data[key]  # type: ignore
        elif isinstance(value, dict) and isinstance(new_data.get(key, None), dict):
            for a, b in value.items():
                if isinstance(b, list):
                    new_data[key][a] = new_data[key][a] + b  # type: ignore
        else:
            new_data[key] = value  # type: ignore

    return new_data


def get_library_path(name: str, path: str | os.PathLike) -> str:
    """
    Returns the path from a library name
    """
    libpath = os.path.join(path, "libraries")
    parts = name.split(":")
    base_path, libname, version = parts[0:3]
    for i in base_path.split("."):
        libpath = os.path.join(libpath, i)
    try:
        version, fileend = version.split("@")
    except ValueError:
        fileend = "jar"

    # construct a filename with the remaining parts
    filename = (
        f"{libname}-{version}{''.join(map(lambda p: f'-{p}', parts[3:]))}.{fileend}"
    )
    libpath = os.path.join(libpath, libname, version, filename)
    return libpath


async def get_jar_mainclass(path: str) -> str:
    """
    Returns the mainclass of a given jar
    """
    zf = zipfile.ZipFile(path)
    # Parse the MANIFEST.MF
    with zf.open("META-INF/MANIFEST.MF") as f:
        lines = f.read().decode("utf-8").splitlines()
    zf.close()
    content = {}
    for i in lines:
        try:
            key, value = i.split(":")
            content[key] = value[1:]
        except Exception:
            pass
    return content["Main-Class"]


async def get_sha1_hash(path: str) -> str:
    """
    Calculate the sha1 checksum of a file
    Source: https://stackoverflow.com/questions/22058048/hashing-a-file-in-python
    """
    BUF_SIZE = 65536
    sha1 = hashlib.sha1()
    async with aiofiles.open(path, "rb") as f:
        while True:
            data = await f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def get_os_version() -> str:
    """
    Try to implement System.getProperty("os.version") from Java for use in rules
    This doesn't work on mac yet
    """
    if platform.system() == "Windows":
        ver = sys.getwindowsversion()  # type: ignore
        return f"{ver.major}.{ver.minor}"
    elif platform.system == "Darwin":
        return ""
    else:
        return platform.uname().release


_user_agent_cache = None


async def get_user_agent() -> str:
    """
    Returns the user agent of minecraft-launcher-lib
    """
    global _user_agent_cache
    if _user_agent_cache is not None:
        return _user_agent_cache
    else:
        file_path = os.path.join(os.path.dirname(__file__), "version.txt")
        async with aiofiles.open(file_path, "r", encoding="utf-8") as f:
            _user_agent_cache = "minecraft-launcher-lib/" + (await f.read()).strip()
            return _user_agent_cache


def get_classpath_separator() -> Literal[":", ";"]:
    """
    Returns the classpath separator for the current os
    """
    if platform.system() == "Windows":
        return ";"
    else:
        return ":"


_requests_response_cache: dict[str, RequestsResponseCache] = {}


async def get_requests_response_cache(url: str) -> aiohttp.ClientResponse:
    """
    Caches the result of request.get(). If a request was made to the same URL within the last hour, the cache will be used, so you don't need to make a request to a URl each timje you call a function.
    """
    now = datetime.datetime.now()

    if (
        url not in _requests_response_cache
        or (now - _requests_response_cache[url]["datetime"]).total_seconds() / 60 / 60
        >= 1
    ):
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url, headers={"user-agent": await get_user_agent()}
            ) as r:
                if r.status == 200:
                    # 將響應體複製出來避免連接關閉後無法訪問
                    content = await r.read()
                    text = await r.text()
                    json_data = (
                        await r.json()
                        if "application/json" in r.headers.get("Content-Type", "")
                        else None
                    )

                    # 創建一個模擬的 response 物件來存儲在緩存中
                    response_cache = {
                        "status": r.status,
                        "content": content,
                        "text": text,
                        "json_data": json_data,
                        "headers": dict(r.headers),
                    }

                    _requests_response_cache[url] = {
                        "response": response_cache,
                        "datetime": now,
                    }
                    return response_cache
                return {
                    "status": r.status,
                    "content": await r.read(),
                    "text": await r.text(),
                    "json_data": None,
                    "headers": dict(r.headers),
                }
    else:
        return _requests_response_cache[url]["response"]


async def parse_maven_metadata(url: str) -> MavenMetadata:
    """
    Parses a maven metadata file
    """
    r = await get_requests_response_cache(url)
    # The structure of the metadata file is simple. So you don't need a XML parser. It can be parsed using RegEx.
    text = r["text"]
    return {
        "release": re.search("(?<=<release>).*?(?=</release>)", text, re.MULTILINE).group(),  # type: ignore
        "latest": re.search("(?<=<latest>).*?(?=</latest>)", text, re.MULTILINE).group(),  # type: ignore
        "versions": re.findall("(?<=<version>).*?(?=</version>)", text, re.MULTILINE),
    }


async def extract_file_from_zip(
    handler: zipfile.ZipFile,
    zip_path: str,
    extract_path: str,
    minecraft_directory: str | os.PathLike | None = None,
) -> None:
    """
    Extract a file from a zip handler into the given path
    """
    if minecraft_directory is not None:
        check_path_inside_minecraft_directory(minecraft_directory, extract_path)

    try:
        os.makedirs(os.path.dirname(extract_path), exist_ok=True)
    except Exception:
        pass

    with handler.open(zip_path, "r") as f:
        data = f.read()
        async with aiofiles.open(extract_path, "wb") as w:
            await w.write(data)


def assert_func(expression: bool) -> None:
    """
    The assert keyword is not available when running Python in Optimized Mode.
    This function is a drop-in replacement.
    See https://docs.python.org/3/using/cmdline.html?highlight=pythonoptimize#cmdoption-O
    """
    if not expression:
        raise AssertionError()


async def get_client_json(
    version: str, minecraft_directory: str | os.PathLike
) -> ClientJson:
    """Load the client.json for the given version"""
    local_path = os.path.join(
        minecraft_directory, "versions", version, f"{version}.json"
    )
    if os.path.isfile(local_path):
        async with aiofiles.open(local_path, "r", encoding="utf-8") as f:
            data = json.loads(await f.read())

        if "inheritsFrom" in data:
            data = await inherit_json(data, minecraft_directory)

        return data

    version_list = (
        await get_requests_response_cache(
            "https://launchermeta.mojang.com/mc/game/version_manifest_v2.json"
        )
    )["json_data"]

    for i in version_list["versions"]:
        if i["id"] == version:
            return (await get_requests_response_cache(i["url"]))["json_data"]

    raise VersionNotFound(version)



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
# This file is part of asyncio-minecraft-launcher-lib (https://github.com/JaydenChao101/asyncio-mc-lancher-lib)
# Copyright (c) 2025 JaydenChao101 <jaydenchao@proton.me> and contributors
# This part of the code is licensed under the MIT License.
from launcher_core.microsoft_types import (
    AuthorizationTokenResponse,
    XBLResponse,
    XSTSResponse,
    MinecraftAuthenticateResponse,
)
from launcher_core.exceptions import (
    AccountBanFromXbox,
    AccountNeedAdultVerification,
    AccountNotHaveXbox,
    XboxLiveNotAvailable,
)
import urllib.parse
import aiohttp
from launcher_core.logging_utils import logger
import asyncio

__AUTH_URL__ = "https://login.live.com/oauth20_authorize.srf"
__TOKEN_URL__ = "https://login.live.com/oauth20_token.srf"
__DEVICE_TOKEN_URL__ = "https://login.microsoftonline.com/consumers/oauth2/v2.0/token"
__SCOPE__ = "XboxLive.signin offline_access"
__DEVICE_CODE_URL__ = (
    "https://login.microsoftonline.com/consumers/oauth2/v2.0/devicecode"
)


class Login:
    def __init__(
        self,
        CLIENT_ID: str = "00000000402b5328",
        REDIRECT_URI: str = "https://login.live.com/oauth20_desktop.srf",
    ):
        self.CLIENT_ID = CLIENT_ID
        self.REDIRECT_URI = REDIRECT_URI

    async def get_login_url(self) -> str:
        """
        Generate a login url.
        :return: The url to the website on which the user logs in
        """
        parameters = {
            "client_id": self.CLIENT_ID,
            "response_type": "code",
            "redirect_uri": self.REDIRECT_URI,
            "response_mode": "query",
            "scope": __SCOPE__,
        }
        url = (
            urllib.parse.urlparse(__AUTH_URL__)
            ._replace(query=urllib.parse.urlencode(parameters))
            .geturl()
        )
        logger.info(f"Login url: {url}")  # Re-enabled logging of login URL.
        return url

    async def extract_code_from_url(self, url: str) -> str:
        """
        Extract the code from the redirect url.

        :param url: The redirect url
        :return: The code
        """
        parsed_url = urllib.parse.urlparse(url)
        query_params = urllib.parse.parse_qs(parsed_url.query)
        if "code" not in query_params:
            raise ValueError("No code found in the URL")
        return query_params["code"][0]

    async def get_ms_token(
        self,
        code: str,
        CLIENT_SECRET: str = None,
        REDIRECT_URI: str = "https://login.live.com/oauth20_desktop.srf",
    ) -> AuthorizationTokenResponse:
        """
        Get the Microsoft token using the code from the login url.

        :param code: The code from the login url
        :return: The Microsoft token
        """
        data = {
            "client_id": self.CLIENT_ID,
            "code": code,
            "grant_type": "authorization_code",
            "redirect_uri": REDIRECT_URI,
            "scope": __SCOPE__,
        }
        if CLIENT_SECRET:
            data["client_secret"] = CLIENT_SECRET
        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                __TOKEN_URL__, data=urllib.parse.urlencode(data), headers=headers
            ) as resp:
                resp.raise_for_status()
                data = await resp.json()
                logger.info(f"Microsoft token response: {data}")
                return data

    async def get_xbl_token(ms_access_token: str) -> XBLResponse:
        payload = {
            "Properties": {
                "AuthMethod": "RPS",
                "SiteName": "user.auth.xboxlive.com",
                "RpsTicket": f"d={ms_access_token}",
            },
            "RelyingParty": "http://auth.xboxlive.com",  # 注意这里
            "TokenType": "JWT",
        }
        headers = {"Content-Type": "application/json"}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                "https://user.auth.xboxlive.com/user/authenticate",
                json=payload,
                headers=headers,
            ) as resp:
                resp.raise_for_status()
                data = await resp.json()
                logger.info(f"Xbox Token response: {data}")
                return data

    async def get_xsts_token(xbl_token: str) -> XSTSResponse:
        """
        Get the XSTS token using the Xbox Live token.

        :param xbl_token: The Xbox Live token
        :return: The XSTS token
        """
        payload = {
            "Properties": {"SandboxId": "RETAIL", "UserTokens": [xbl_token]},
            "RelyingParty": "rp://api.minecraftservices.com/",
            "TokenType": "JWT",
        }
        headers = {"Content-Type": "application/json"}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                "https://xsts.auth.xboxlive.com/xsts/authorize",
                json=payload,
                headers=headers,
            ) as resp:
                status = resp.status
                data = await resp.json()

                if status == 401:
                    error_code = data.get("XErr")
                    """
                    The Redirect parameter usually will not resolve or go anywhere in a browser, likely they're targeting Xbox consoles.

                    Noted XErr codes and their meanings:

                    2148916227: The account is banned from Xbox.
                    2148916233: The account doesn't have an Xbox account. Once they sign up for one (or login through minecraft.net to create one) then they can proceed with the login. This shouldn't happen with accounts that have purchased Minecraft with a Microsoft account, as they would've already gone through that Xbox signup process.
                    2148916235: The account is from a country where Xbox Live is not available/banned
                    2148916236: The account needs adult verification on Xbox page. (South Korea)
                    2148916237: The account needs adult verification on Xbox page. (South Korea)
                    2148916238: The account is a child (under 18) and cannot proceed unless the account is added to a Family by an adult. This only seems to occur when using a custom Microsoft Azure application. When using the Minecraft launchers client id, this doesn't trigger.
                    """
                    error_codes = {
                        "ban_from_Xbox": 2148916227,
                        "didnt_have_xbox": 2148916233,
                        "ban_contry": 2148916235,
                        "need_adult_verification": 2148916236,
                        "need_adult_verification_1": 2148916237,
                    }

                    # raise Error
                    if error_code == error_codes["ban_from_Xbox"]:
                        raise AccountBanFromXbox()
                    elif error_code == error_codes["didnt_have_xbox"]:
                        raise AccountNotHaveXbox()
                    elif error_code == error_codes["ban_contry"]:
                        raise XboxLiveNotAvailable()
                    elif (
                        error_code == error_codes["need_adult_verification"]
                        or error_code == error_codes["need_adult_verification_1"]
                    ):
                        raise AccountNeedAdultVerification()
                    else:
                        raise Exception(
                            f"Loginning of the XSTS token error: {error_code}"
                        )

                logger.info(f"XSTS Token response: {data}")
                return data

    async def get_minecraft_access_token(
        xsts_token: str, uhs: str
    ) -> MinecraftAuthenticateResponse:
        """
        Get the Minecraft access token using the XSTS token and user hash.

        :param xsts_token: The XSTS token
        :param uhs: The user hash
        :return: The Minecraft access token
        """
        identity_token = f"XBL3.0 x={uhs};{xsts_token}"
        payload = {"identityToken": identity_token}
        headers = {"Content-Type": "application/json"}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                "https://api.minecraftservices.com/authentication/login_with_xbox",
                json=payload,
                headers=headers,
            ) as resp:
                resp.raise_for_status()
                data = await resp.json()
                logger.info(f"Minecraft access token response: {data}")
                return data


class device_code_login:
    def __init__(
        self,
        CLIENT_ID: str,
    ):
        self.CLIENT_ID = CLIENT_ID

    async def get_device_code(self) -> dict:
        """
        Get the device code using the client id and redirect uri.

        :return: The device code
        """
        data = {
            "client_id": self.CLIENT_ID,
            "scope": __SCOPE__,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                __DEVICE_CODE_URL__, data=data, headers=headers
            ) as resp:
                resp.raise_for_status()
                return await resp.json()

    async def poll_device_code(self, device_code: str, interval: int, expires_in: int):
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "client_id": self.CLIENT_ID,
            "device_code": device_code,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        elapsed = 0
        while elapsed < expires_in:
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    __DEVICE_TOKEN_URL__, data=data, headers=headers
                ) as resp:
                    result = await resp.json()
                    if "access_token" in result:
                        return result
                    elif result.get("error") == "authorization_pending":
                        await asyncio.sleep(interval)
                        elapsed += interval
                    elif result.get("error") == "slow_down":
                        interval += 5
                        await asyncio.sleep(interval)
                        elapsed += interval
                    else:
                        raise Exception(f"Device code flow error: {result}")
        raise Exception("Device code expired or not authorized in time.")


async def refresh_minecraft_token(
    refresh_token: str,
    CLIENT_ID: str = "00000000402b5328",
    CLIENT_SECRET: str = None,
) -> AuthorizationTokenResponse:
    """
    Refresh the Minecraft token using the refresh token.

    :param refresh_token: The refresh token
    :return: The refreshed Minecraft token
    """
    data = {
        "client_id": CLIENT_ID,
        "refresh_token": refresh_token,
        "grant_type": "refresh_token",
        "scope": __SCOPE__,
    }
    if CLIENT_SECRET:
        data["client_secret"] = CLIENT_SECRET
    headers = {"Content-Type": "application/x-www-form-urlencoded"}

    async with aiohttp.ClientSession() as session:
        async with session.post(
            __TOKEN_URL__, data=urllib.parse.urlencode(data), headers=headers
        ) as resp:
            resp.raise_for_status()
            data = await resp.json()
            logger.info(f"Refreshed Minecraft token response: {data}")
            return data


# This example shows how to login to a Microsoft account and get the access token.
"""async def login_minecraft():
    import webbrowser
    webbrowser.open_new_tab(await get_login_url())
    url = input("Please open the URL in your browser and paste the redirect URL here: ")
    code = await extract_code_from_url(url)
    ms_token = await get_ms_token(code)
    xbl_token = await get_xbl_token(ms_token["access_token"])
    xsts_token = await get_xsts_token(xbl_token["Token"])
    uhs = xbl_token["DisplayClaims"]["xui"][0]["uhs"]
    minecraft_token = await get_minecraft_access_token(xsts_token["Token"], uhs)
    await have_minecraft(minecraft_token["access_token"])
    return {
        "access_token": minecraft_token["access_token"],
        "refresh_token": ms_token["refresh_token"],
        "expires_in": ms_token["expires_in"],
        "uhs": uhs,
        "xsts_token": xsts_token["Token"],
        "xbl_token": xbl_token["Token"]
    }

if __name__ == "__main__":
    import asyncio
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(login_minecraft())
    print(result)
"""



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause

from .logging_utils import logger

from . import (
    command,
    install,
    microsoft_account,
    utils,
    java_utils,
    forge,
    fabric,
    quilt,
    news,
    runtime,
    vanilla_launcher,
    mrpack,
    exceptions,
    _types,
    microsoft_types,
)

__all__ = [
    "command",
    "install",
    "microsoft_account",
    "utils",
    "news",
    "java_utils",
    "forge",
    "fabric",
    "quilt",
    "runtime",
    "vanilla_launcher",
    "mrpack",
    "exceptions",
    "_types",
    "microsoft_types",
    "logger",
]



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"""
quilt contains functions for dealing with the `Quilt modloader <https://quiltmc.org>`_.

You may have noticed, that the Functions are the same as in the :doc:`fabric` module.
That's because Quilt is a Fork of Fabric. This module behaves exactly the same as the fabric module.
"""
import asyncio
import tempfile
import os

from ._helper import (
    download_file,
    get_requests_response_cache,
    parse_maven_metadata,
    empty,
    SUBPROCESS_STARTUP_INFO,
)
from .exceptions import VersionNotFound, UnsupportedVersion, ExternalProgramError
from ._types import QuiltMinecraftVersion, QuiltLoader, CallbackDict
from .install import install_minecraft_version
from .utils import is_version_valid


async def get_all_minecraft_versions() -> list[QuiltMinecraftVersion]:
    """
    Returns all available Minecraft Versions for Quilt

    Example:

    .. code:: python

        for version in await launcher_corequilt.get_all_minecraft_versions():
            print(version["version"])
    """
    quilt_minecraft_versions_url = "https://meta.quiltmc.org/v3/versions/game"
    return await get_requests_response_cache(quilt_minecraft_versions_url)


async def get_stable_minecraft_versions() -> list[str]:
    """
    Returns a list which only contains the stable Minecraft versions that supports Quilt

    Example:

    .. code:: python

        for version in await launcher_corequilt.get_stable_minecraft_versions():
            print(version)
    """
    minecraft_versions = await get_all_minecraft_versions()
    stable_versions = []
    for i in minecraft_versions:
        if i["stable"] is True:
            stable_versions.append(i["version"])
    return stable_versions


async def get_latest_minecraft_version() -> str:
    """
    Returns the latest unstable Minecraft versions that supports Quilt. This could be a snapshot.

    Example:

    .. code:: python

        print("Latest Minecraft version: " +
            await launcher_corequilt.get_latest_minecraft_version())
    """
    minecraft_versions = await get_all_minecraft_versions()
    return minecraft_versions[0]["version"]


async def get_latest_stable_minecraft_version() -> str:
    """
    Returns the latest stable Minecraft version that supports Quilt

    Example:

    .. code:: python

        print("Latest stable Minecraft version: " +
            await launcher_corequilt.get_latest_stable_minecraft_version())
    """
    stable_versions = await get_stable_minecraft_versions()
    return stable_versions[0]


async def is_minecraft_version_supported(version: str) -> bool:
    """
    Checks if a Minecraft version supported by Quilt

    Example:

    .. code:: python

        version = "1.20"
        if await launcher_corequilt.is_minecraft_version_supported(version):
            print(f"{version} is supported by quilt")
        else:
            print(f"{version} is not supported by quilt")

    :param version: A vanilla version
    """
    minecraft_versions = await get_all_minecraft_versions()
    for i in minecraft_versions:
        if i["version"] == version:
            return True
    return False


async def get_all_loader_versions() -> list[QuiltLoader]:
    """
    Returns all loader versions

    Example:

    .. code:: python

        for version in await launcher_corequilt.get_all_loader_versions():
            print(version["version"])
    """
    quilt_loader_versions_url = "https://meta.quiltmc.org/v3/versions/loader"
    return await get_requests_response_cache(quilt_loader_versions_url)


async def get_latest_loader_version() -> str:
    """
    Get the latest loader version


    Example:

    .. code:: python

        print("Latest loader version: " + await launcher_corequilt.get_latest_loader_version())
    """
    loader_versions = await get_all_loader_versions()
    return loader_versions[0]["version"]


async def get_latest_installer_version() -> str:
    """
    Returns the latest installer version

    Example:

    .. code:: python

        print("Latest installer version: " +
            await launcher_corequilt.get_latest_installer_version())
    """
    quilt_installer_maven_url = "https://maven.quiltmc.org/repository/release/org/quiltmc/quilt-installer/maven-metadata.xml"
    result = await parse_maven_metadata(quilt_installer_maven_url)
    return result["latest"]


async def install_quilt(
    minecraft_version: str,
    minecraft_directory: str | os.PathLike,
    loader_version: str | None = None,
    callback: CallbackDict | None = None,
    java: str | os.PathLike | None = None,
) -> None:
    """
    Installs the Quilt modloader.

    Example:

    .. code:: python

        minecraft_version = "1.20"
        minecraft_directory = launcher_coreutils.get_minecraft_directory()
        await launcher_corequilt.install_quilt(minecraft_version, minecraft_directory)

    :param minecraft_version: A vanilla version that is supported by Quilt
    :param minecraft_directory: The path to your Minecraft directory
    :param loader_version: The Quilt loader version. If not given it will use the latest
    :param callback: The same dict as for :func:`~launcher_coreinstall.install_minecraft_version`
    :param java: A Path to a custom Java executable
    :raises VersionNotFound: The given Minecraft does not exists
    :raises UnsupportedVersion: The given Minecraft version is not supported by Quilt
    """
    path = str(minecraft_directory)
    if not callback:
        callback = {}

    # Check if the given version exists
    if not await is_version_valid(minecraft_version, minecraft_directory):
        raise VersionNotFound(minecraft_version)

    # Check if the given Minecraft version supported
    if not await is_minecraft_version_supported(minecraft_version):
        raise UnsupportedVersion(minecraft_version)

    # Get latest loader version if not given
    if not loader_version:
        loader_version = await get_latest_loader_version()

    # Make sure the Minecraft version is installed
    await install_minecraft_version(minecraft_version, path, callback=callback)

    # Get installer version
    installer_version = await get_latest_installer_version()
    installer_download_url = (
        f"https://maven.quiltmc.org/repository/release/org/quiltmc/"
        f"quilt-installer/{installer_version}/quilt-installer-{installer_version}.jar"
    )

    with tempfile.TemporaryDirectory(
        prefix="minecraft-launcher-lib-quilt-install-"
    ) as tempdir:
        installer_path = os.path.join(tempdir, "quit-installer.jar")

        # Download the installer
        await download_file(
            installer_download_url, installer_path, callback=callback, overwrite=True
        )

        # Run the installer
        callback.get("setStatus", empty)("Running quilt installer")
        command = [
            "java" if java is None else str(java),
            "-jar",
            installer_path,
            "install",
            "client",
            minecraft_version,
            loader_version,
            f"--install-dir={path}",
            "--no-profile",
        ]

        process = await asyncio.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            startupinfo=SUBPROCESS_STARTUP_INFO,
        )

        stdout, stderr = await process.communicate()

        if process.returncode != 0:
            raise ExternalProgramError(command, stdout, stderr)

    # Install all libs of quilt
    quilt_minecraft_version = f"quilt-loader-{loader_version}-{minecraft_version}"
    await install_minecraft_version(quilt_minecraft_version, path, callback=callback)



"""
The file is part of the minecraft_launcher_lib project.
It is licensed under the BSD-2-Clause license.
It contains functions for interacting with the Minecraft skin API.
It provides functions to get the skin URL, get the skin data, and download the skin.
Copyright (c) 2025 JaydenChao101 <jaydenchao@proton.me> and contributors
"""

from ._types import Skin, Credential, MinecraftProfileResponse
from .exceptions import AccountNotOwnMinecraft
from base64 import b64decode
import json
from typing import Optional
import aiohttp
import aiofiles


class Skin:

    def __init__(self, Credential: Credential):
        self.Credential = Credential

    async def get_skin_and_cape(self) -> Optional[Skin]:
        """
        取得指定 UUID 的外觀與披風 URL。

        :param Credential: 玩家凭证，需包含 uuid
        :return: dict，包含 skin 和 cape（如无披風则 cape 为 None）
        """
        Credential = self.Credential
        uuid = Credential.uuid
        url = f"https://sessionserver.mojang.com/session/minecraft/profile/{uuid}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    return None
                data = await response.json()
        properties = data.get("properties", [])
        for prop in properties:
            if prop.get("name") == "textures":
                value = prop.get("value")
                decoded = b64decode(value).decode("utf-8")
                textures_json = json.loads(decoded)
                textures = textures_json.get("textures", {})
                skin_url = textures.get("SKIN", {}).get("url")
                cape_url = textures.get("CAPE", {}).get("url")
                return {"skin": skin_url, "cape": cape_url}
        return None

    async def change_skin(self, skin_url: str, model: str = "") -> bool:
        """
        更改玩家的外觀（通过URL）。

        :param Credential: 玩家凭证，需包含 uuid 和 access_token
        :param skin_url: 新的外觀 URL
        :param model: "slim" 或 ""，默认为 ""
        :param cape_url: 未使用，保留参数
        :return: bool，是否成功更改
        """
        Credential = self.Credential
        uuid = Credential.uuid
        access_token = Credential.access_token
        url = f"https://api.mojang.com/user/profile/{uuid}/skin"
        headers = {"Authorization": f"Bearer {access_token}"}
        data = {"model": model, "url": skin_url}
        async with aiohttp.ClientSession() as session:
            async with session.post(url, headers=headers, json=data) as response:
                return response.status == 204

    async def upload_skin(self, file_path: str, model: str = "") -> bool:
        """
        上传本地外观文件。

        :param Credential: 玩家凭证，需包含 uuid 和 access_token
        :param file_path: 本地皮肤文件路径
        :param model: "slim" 或 ""，默认为 ""
        :return: bool，是否成功上传
        """
        Credential = self.Credential
        uuid = Credential.uuid
        access_token = Credential.access_token
        url = f"https://api.mojang.com/user/profile/{uuid}/skin"
        headers = {"Authorization": f"Bearer {access_token}"}

        async with aiofiles.open(file_path, "rb") as f:
            file_data = await f.read()

        form = aiohttp.FormData()
        form.add_field("file", file_data, filename=file_path.split("/")[-1])
        form.add_field("model", model)

        async with aiohttp.ClientSession() as session:
            async with session.put(url, headers=headers, data=form) as response:
                return response.status == 204

    async def reset_skin(self) -> bool:
        """
        重设玩家外观为默认。

        :param Credential: 玩家凭证，需包含 uuid 和 access_token
        :return: bool，是否成功重设
        """
        Credential = self.Credential
        uuid = Credential.uuid
        access_token = Credential.access_token
        url = f"https://api.mojang.com/user/profile/{uuid}/skin"
        headers = {"Authorization": f"Bearer {access_token}"}
        async with aiohttp.ClientSession() as session:
            async with session.delete(url, headers=headers) as response:
                return response.status == 204


async def have_minecraft(access_token: str) -> bool:
    """
    Check if the user owns Minecraft using the access token.

    :param access_token: The Minecraft access token
    :return: True if the user owns Minecraft, Raise AccountNotOwnMinecraft otherwise
    """
    headers = {"Authorization": f"Bearer {access_token}"}

    async with aiohttp.ClientSession() as session:
        async with session.get(
            "https://api.minecraftservices.com/entitlements/mcstore", headers=headers
        ) as resp:
            resp.raise_for_status()
            data = await resp.json()
            if not data.get("items"):
                raise AccountNotOwnMinecraft()
            return True


async def get_minecraft_profile(access_token: str) -> MinecraftProfileResponse:
    """
    Get the Minecraft profile using the access token.

    :param access_token: The Minecraft access token
    :return: The Minecraft profile
    """
    headers = {"Authorization": f"Bearer {access_token}"}

    async with aiohttp.ClientSession() as session:
        async with session.get(
            "https://api.minecraftservices.com/minecraft/profile", headers=headers
        ) as resp:
            resp.raise_for_status()
            return await resp.json()


async def get_minecraft_player_attributes(
    access_token: str,
) -> MinecraftProfileResponse:
    """
    Get the Minecraft player attributes.

    :return: The Minecraft player attributes
    """
    headers = {"Authorization": f"Bearer {access_token}"}

    async with aiohttp.ClientSession() as session:
        async with session.get(
            "https://api.minecraftservices.com/minecraft/profile/attributes",
            headers=headers,
        ) as resp:
            resp.raise_for_status()
            return await resp.json()



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
from typing import TypedDict


class _AssetsJsonObject(TypedDict):
    hash: str
    size: int


class AssetsJson(TypedDict):
    objects: dict[str, _AssetsJsonObject]



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
from .shared_types import ClientJson, ClientJsonLibrary
from typing import Literal, TypedDict


class _ForgeInstallProcessor(TypedDict, total=False):
    sides: list[Literal["client", "server"]]
    jar: str
    classpath: list[str]
    args: list[str]


class _ForgeInstallProfileInstall(TypedDict, total=False):
    profileName: str
    target: str
    path: str
    version: str
    filePath: str
    welcome: str
    minecraft: str
    mirrorList: str
    logo: str


class ForgeInstallProfile(TypedDict, total=False):
    spec: int
    profile: str
    version: str
    minecraft: str
    serverJarPath: str
    data: dict[str, dict[Literal["client", "server"], str]]
    processors: list[_ForgeInstallProcessor]
    libraries: list[ClientJsonLibrary]
    icon: str
    logo: str
    mirrorList: str
    welcome: str
    install: _ForgeInstallProfileInstall
    versionInfo: ClientJson



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
from typing import Union, Literal, TypedDict


class ClientJsonRule(TypedDict):
    action: Literal["allow", "disallow"]
    os: dict[Literal["name", "arch", "vesion"], str]
    features: dict[
        Literal[
            "has_custom_resolution",
            "is_demo_user",
            "has_quick_plays_support",
            "is_quick_play_singleplayer",
            "is_quick_play_multiplayer",
            "is_quick_play_realms",
        ],
        bool,
    ]


class ClientJsonArgumentRule(TypedDict, total=False):
    compatibilityRules: list[ClientJsonRule]
    rules: list[ClientJsonRule]
    value: Union[str, list[str]]


class _ClientJsonAssetIndex(TypedDict):
    id: str
    sha1: str
    size: int
    totalSize: int
    url: str


class _ClientJsonDownloads(TypedDict):
    sha1: str
    size: int
    url: str


class _ClientJsonJavaVersion(TypedDict):
    component: str
    majorVersion: int


class _ClientJsonLibraryDownloadsArtifact(TypedDict):
    path: str
    url: str
    sha1: str
    size: int


class _ClientJsonLibraryDownloads(TypedDict, total=False):
    artifact: _ClientJsonLibraryDownloadsArtifact
    classifiers: dict[
        Literal[
            "javadoc", "natives-linux", "natives-macos", "natives-windows", "sources"
        ],
        _ClientJsonLibraryDownloadsArtifact,
    ]


class ClientJsonLibrary(TypedDict, total=False):
    name: str
    downloads: _ClientJsonLibraryDownloads
    extract: dict[Literal["exclude"], list[str]]
    rules: list[ClientJsonRule]
    natives: dict[Literal["linux", "osx", "windows"], str]
    url: str


class _ClientJsonLoggingFile(TypedDict):
    id: str
    sha1: str
    size: int
    url: str


class _ClientJsonLogging(TypedDict):
    argument: str
    file: _ClientJsonLoggingFile
    type: str


class ClientJson(TypedDict, total=False):
    id: str
    jar: str
    arguments: dict[Literal["game", "jvm"], list[Union[str, ClientJsonArgumentRule]]]
    minecraftArguments: str
    assetIndex: _ClientJsonAssetIndex
    assets: str
    downloads: dict[
        Literal["client", "client_mappings", "server", "server_mappings"],
        _ClientJsonDownloads,
    ]
    javaVersion: _ClientJsonJavaVersion
    libraries: list[ClientJsonLibrary]
    logging: dict[Literal["client"], _ClientJsonLogging]
    mainClass: str
    minimumLauncherVersion: int
    releaseTime: str
    time: str
    type: str
    complianceLevel: int
    inheritsFrom: str


class _VersionListManifestJsonVersion(TypedDict):
    id: str
    type: str
    url: str
    time: str
    releaseTime: str
    sha1: str
    complianceLevel: int


class VersionListManifestJson(TypedDict):
    latest: dict[Literal["release", "snapshot"], str]
    versions: list[_VersionListManifestJsonVersion]



"java_utils contains some functions to help with Java"

from ._helper import SUBPROCESS_STARTUP_INFO
from ._types import JavaInformation
import asyncio
import platform
import re
import os


async def get_java_information(path: str | os.PathLike) -> JavaInformation:
    """
    Returns Some Information about the given Java Installation

    .. note::
        This Function executes the Java executable to determine details such as the version. This might be a security risk.

    Example:

    .. code:: python

        java_path = "<path>"
        information = await launcher_corejava_utils.get_java_information(java_path)
        print("Name: " + information["name"])
        print("Version: " + information["version"])
        print("Java path: " + information["java_path"])

    :param path: The Path to the Installation. It must be the Directory. If your Java executable is e.g. /usr/lib/jvm/java-19-openjdk-amd64/bin/java this Parameter must be /usr/lib/jvm/java-19-openjdk-amd64.
    :return: A dict with Information about the given java installation
    :raises ValueError: Wrong path
    """
    if platform.system() == "Windows":
        if not os.path.isfile(os.path.join(path, "bin", "java.exe")):
            raise ValueError(
                os.path.abspath(os.path.join(path, "bin", "java.exe"))
                + " was not found"
            )
    else:
        if not os.path.isfile(os.path.join(path, "bin", "java")):
            raise ValueError(
                os.path.abspath(os.path.join(path, "bin", "java")) + " was not found"
            )

    process = await asyncio.create_subprocess_exec(
        os.path.join(path, "bin", "java"),
        "-showversion",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        startupinfo=SUBPROCESS_STARTUP_INFO,
    )

    _, stderr = await process.communicate()
    lines = stderr.decode("utf-8").splitlines()

    information: JavaInformation = {}  # type: ignore
    information["path"] = str(path)
    information["name"] = os.path.basename(path)
    information["version"] = re.search(r'(?<=version ")[\d|\.|_]+(?=")', lines[0]).group()  # type: ignore
    information["is_64bit"] = "64-Bit" in lines[2]
    information["openjdk"] = lines[0].startswith("openjdk")

    if platform.system() == "Windows":
        information["java_path"] = os.path.join(
            os.path.abspath(path), "bin", "java.exe"
        )
        information["javaw_path"] = os.path.join(
            os.path.abspath(path), "bin", "javaw.exe"
        )
    else:
        information["java_path"] = os.path.join(os.path.abspath(path), "bin", "java")
        information["javaw_path"] = None

    return information


def _search_java_directory(path: str | os.PathLike) -> list[str]:
    "Used by find_system_java_versions() to parse a Directory"
    if not os.path.isdir(path):
        return []

    java_list: list[str] = []
    for i in os.listdir(path):
        current_entry = os.path.join(path, i)

        if os.path.isfile(current_entry) or os.path.islink(current_entry):
            continue

        if os.path.isfile(os.path.join(current_entry, "bin", "java")) or os.path.isfile(
            os.path.join(current_entry, "bin", "java.exe")
        ):
            java_list.append(current_entry)

    return java_list


async def find_system_java_versions(
    additional_directories: list[str | os.PathLike] | None = None,
) -> list[str]:
    """
    Try to find all Java Versions installed on the System. You can use this to e.g. let the User choose between different Java Versions in a Dropdown.
    macOS is not supported yet.

    Example:

    .. code:: python

        for version in await launcher_corejava_utils.find_system_java_versions():
            print(version)

    :param additional_directories: A List of additional Directories to search for Java in custom locations
    :return: A List with all Directories of Java Installations
    """
    java_list: list[str] = []

    match platform.system():
        case "Windows":
            java_list += _search_java_directory(r"C:\Program Files (x86)\Java")
            java_list += _search_java_directory(r"C:\Program Files\Java")
        case "Linux":
            java_list += _search_java_directory("/usr/lib/jvm")
            java_list += _search_java_directory("/usr/lib/sdk")

    if additional_directories is not None:
        for i in additional_directories:
            java_list += _search_java_directory(i)

    return java_list


async def find_system_java_versions_information(
    additional_directories: list[str | os.PathLike] | None = None,
) -> list[JavaInformation]:
    """
    Same as :func:`find_system_java_version`, but uses :func:`get_java_information` to get some Information about the Installation instead of just proving a Path.
    macOS is not supported yet

    .. note::
        This Function executes the Java executable to determine details such as the version. This might be a security risk.

    Example:

    .. code:: python

        for version_information in await launcher_corejava_utils.find_system_java_versions_information():
            print("Path: " + version_information["path"])
            print("Name: " + version_information["name"])
            print("Version: " + version_information["version"])
            print("Java path: " + version_information["java_path"])
            print()

    :param additional_directories: A List of additional Directories to search for Java in custom locations
    :return: A List with Information of Java Installations
    """
    java_information_list: list[JavaInformation] = []
    java_versions = await find_system_java_versions(
        additional_directories=additional_directories
    )

    for i in java_versions:
        java_information_list.append(await get_java_information(i))

    return java_information_list



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
from typing import Literal, TypedDict


class _RuntimeListJsonEntryManifest(TypedDict):
    sha1: str
    size: int
    url: str


class _RuntimeListJsonEntry(TypedDict):
    availability: dict[Literal["group", "progress"], int]
    manifest: _RuntimeListJsonEntryManifest
    version: dict[Literal["name", "released"], str]


RuntimeListJson = dict[str, dict[str, list[_RuntimeListJsonEntry]]]


class _PlatformManifestJsonFileDownloads(TypedDict):
    sha1: str
    size: int
    url: str


class _PlatformManifestJsonFile(TypedDict, total=False):
    downloads: dict[Literal["lzma", "raw"], _PlatformManifestJsonFileDownloads]
    type: Literal["file", "directory", "link"]
    executable: bool
    target: str


class PlatformManifestJson(TypedDict):
    files: dict[str, _PlatformManifestJsonFile]



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"""
mrpack allows you to install Modpacks from the `Mrpack Format <https://support.modrinth.com/en/articles/8802351-modrinth-modpack-format-mrpack>`_.
You should also take a look at the :doc:`complete example </examples/Mrpack>`.
"""

# 標準庫導入
import zipfile
import json
import os

# 第三方庫導入
import aiohttp
import aiofiles

# 本地導入
from ._helper import download_file, empty, check_path_inside_minecraft_directory
from ._types import MrpackInformation, MrpackInstallOptions, CallbackDict
from ._internal_types.mrpack_types import MrpackIndex, MrpackFile
from .install import install_minecraft_version
from .forge import install_forge_version
from .exceptions import VersionNotFound
from .fabric import install_fabric
from .quilt import install_quilt


def _filter_mrpack_files(
    file_list: list[MrpackFile], mrpack_install_options: MrpackInstallOptions
) -> list[MrpackFile]:
    """
    Gets all Mrpack Files that should be installed
    """
    filtered_list: list[MrpackFile] = []
    for file in file_list:
        if "env" not in file:
            filtered_list.append(file)
            continue

        if file["env"]["client"] == "required":
            filtered_list.append(file)
        if file["env"]["client"] == "optional" and file[
            "path"
        ] in mrpack_install_options.get("optionalFiles", []):
            filtered_list.append(file)

    return filtered_list


async def get_mrpack_information(path: str | os.PathLike) -> MrpackInformation:
    """
    Gets some Information from a .mrpack file

    Example:

    .. code:: python

        path = "/path/to/mrpack"
        information = await launcher_coremrpack.get_mrpack_information(path)
        print("Name: " + information["name"])
        print("Summary: " + information["summary"])

    :param path: The Path the the .mrpack file
    :type path: Union[str, os.PathLike]
    :return: The Information about the given Mrpack
    """
    with zipfile.ZipFile(path, "r") as zf:
        with zf.open("modrinth.index.json", "r") as f:
            index_data = f.read()
            index: MrpackIndex = json.loads(index_data)

            information: MrpackInformation = {}  # type: ignore
            information["name"] = index["name"]
            information["summary"] = index.get("summary", "")
            information["versionId"] = index["versionId"]
            information["formatVersion"] = index["formatVersion"]
            information["minecraftVersion"] = index["dependencies"]["minecraft"]

            information["optionalFiles"] = []
            for file in index["files"]:
                if "env" not in file:
                    continue

                if file["env"]["client"] == "optional":
                    information["optionalFiles"].append(file["path"])

            return information


async def install_mrpack(
    path: str | os.PathLike,
    minecraft_directory: str | os.PathLike,
    modpack_directory: str | os.PathLike | None = None,
    callback: CallbackDict | None = None,
    mrpack_install_options: MrpackInstallOptions | None = None,
) -> None:
    """
    Installs a .mrpack file

    ``mrpack_install_options`` is a dict. All Options are Optional.

    .. code:: python

        mrpack_install_options = {
            "optionalFiles": [], # List with all Optional files
            "skipDependenciesInstall": False # If you want to skip the Dependencies install. Only used for testing purposes.
        }

    Example:

    .. code:: python

        path = "/path/to/mrpack"
        minecraft_directory = minecraft_directory.utils.get_minecraft_directory()
        await launcher_coremrpack.install_mrpack(path, minecraft_directory)

    :param path: The Path the the .mrpack file
    :param minecraft_directory: he path to your Minecraft directory
    :param modpack_directory: If you want to install the Pack in another Directory than your Minecraft directory, set it here.
    :param callback: The same dict as for :func:`~launcher_coreinstall.install_minecraft_version`
    :param mrpack_install_options: Some Options to install the Pack (see below)
    :raises FileOutsideMinecraftDirectory: A File should be placed outside the given Minecraft directory
    """
    minecraft_directory = os.path.abspath(minecraft_directory)
    path = os.path.abspath(path)

    if modpack_directory is None:
        modpack_directory = minecraft_directory
    else:
        modpack_directory = os.path.abspath(modpack_directory)

    if callback is None:
        callback = {}

    if mrpack_install_options is None:
        mrpack_install_options = {}

    with zipfile.ZipFile(path, "r") as zf:
        with zf.open("modrinth.index.json", "r") as f:
            index_data = f.read()
            index: MrpackIndex = json.loads(index_data)

        # Download the files
        callback.get("setStatus", empty)("Download mrpack files")
        file_list = _filter_mrpack_files(index["files"], mrpack_install_options)
        callback.get("setMax", empty)(len(file_list))
        for count, file in enumerate(file_list):
            full_path = os.path.abspath(os.path.join(modpack_directory, file["path"]))

            check_path_inside_minecraft_directory(modpack_directory, full_path)

            await download_file(
                file["downloads"][0],
                full_path,
                sha1=file["hashes"]["sha1"],
                callback=callback,
            )

            callback.get("setProgress", empty)(count + 1)

        # Extract the overrides
        callback.get("setStatus", empty)("Extract overrides")
        for zip_name in zf.namelist():
            # Check if the entry is in the overrides and if it is a file
            if (
                not zip_name.startswith("overrides/")
                and not zip_name.startswith("client-overrides/")
            ) or zf.getinfo(zip_name).file_size == 0:
                continue

            # Remove the overrides at the start of the Name
            # We don't have removeprefix() in Python 3.8
            if zip_name.startswith("client-overrides/"):
                file_name = zip_name[len("client-overrides/") :]
            else:
                file_name = zip_name[len("overrides/") :]

            # Constructs the full Path
            full_path = os.path.abspath(os.path.join(modpack_directory, file_name))

            check_path_inside_minecraft_directory(modpack_directory, full_path)

            callback.get("setStatus", empty)(f"Extract {zip_name}]")

            try:
                os.makedirs(os.path.dirname(full_path), exist_ok=True)
            except FileExistsError:
                pass

            file_data = zf.read(zip_name)
            async with aiofiles.open(full_path, "wb") as f:
                await f.write(file_data)

        if mrpack_install_options.get("skipDependenciesInstall"):
            return

        # Install dependencies
        callback.get("setStatus", empty)(
            "Installing Minecraft " + index["dependencies"]["minecraft"]
        )
        await install_minecraft_version(
            index["dependencies"]["minecraft"], minecraft_directory, callback=callback
        )

        if "forge" in index["dependencies"]:
            forge_version = None
            FORGE_DOWNLOAD_URL = "https://maven.minecraftforge.net/net/minecraftforge/forge/{version}/forge-{version}-installer.jar"

            # Check forge version availability
            async with aiohttp.ClientSession() as session:
                for current_forge_version in (
                    index["dependencies"]["minecraft"]
                    + "-"
                    + index["dependencies"]["forge"],
                    index["dependencies"]["minecraft"]
                    + "-"
                    + index["dependencies"]["forge"]
                    + "-"
                    + index["dependencies"]["minecraft"],
                ):
                    url = FORGE_DOWNLOAD_URL.format(version=current_forge_version)
                    async with session.head(url) as response:
                        if response.status == 200:
                            forge_version = current_forge_version
                            break
                else:
                    raise VersionNotFound(index["dependencies"]["forge"])

            callback.get("setStatus", empty)(f"Installing Forge {forge_version}")
            await install_forge_version(
                forge_version, minecraft_directory, callback=callback
            )

        if "fabric-loader" in index["dependencies"]:
            callback.get("setStatus", empty)(
                "Installing Fabric "
                + index["dependencies"]["fabric-loader"]
                + " for Minecraft "
                + index["dependencies"]["minecraft"]
            )
            await install_fabric(
                index["dependencies"]["minecraft"],
                minecraft_directory,
                loader_version=index["dependencies"]["fabric-loader"],
                callback=callback,
            )

        if "quilt-loader" in index["dependencies"]:
            callback.get("setStatus", empty)(
                "Installing Quilt "
                + index["dependencies"]["quilt-loader"]
                + " for Minecraft "
                + index["dependencies"]["minecraft"]
            )
            await install_quilt(
                index["dependencies"]["minecraft"],
                minecraft_directory,
                loader_version=index["dependencies"]["quilt-loader"],
                callback=callback,
            )


async def get_mrpack_launch_version(path: str | os.PathLike) -> str:
    """
    Returns that Version that needs to be used with :func:`~launcher_corecommand.get_minecraft_command`.

    Example:

    .. code:: python

        path = "/path/to/mrpack"
        print("Launch version: " + await launcher_coremrpack.get_mrpack_launch_version(path))

    :param path: The Path the the .mrpack file
    :return: The version
    """
    with zipfile.ZipFile(path, "r") as zf:
        with zf.open("modrinth.index.json", "r") as f:
            index_data = f.read()
            index: MrpackIndex = json.loads(index_data)

            if "forge" in index["dependencies"]:
                return (
                    index["dependencies"]["minecraft"]
                    + "-forge-"
                    + index["dependencies"]["forge"]
                )
            if "fabric-loader" in index["dependencies"]:
                return (
                    "fabric-loader-"
                    + index["dependencies"]["fabric-loader"]
                    + "-"
                    + index["dependencies"]["minecraft"]
                )
            if "quilt-loader" in index["dependencies"]:
                return (
                    "quilt-loader-"
                    + index["dependencies"]["quilt-loader"]
                    + "-"
                    + index["dependencies"]["minecraft"]
                )
            return index["dependencies"]["minecraft"]



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"""
This module contains all Types for minecraft-launcher-lib. It may help your IDE. You don't need to use this module directly in your code.
If you are not interested in static typing just ignore it.
For more information about TypedDict see `PEP 589 <https://peps.python.org/pep-0589/>`_.
"""
from typing import Literal, TypedDict, Callable
import datetime
from uuid import UUID
from dataclasses import dataclass


class MinecraftOptions(TypedDict, total=False):
    username: str
    uuid: UUID
    token: str
    executablePath: str
    defaultExecutablePath: str
    jvmArguments: list[str]
    launcherName: str
    launcherVersion: str
    gameDirectory: str
    demo: bool
    customResolution: bool
    resolutionWidth: str
    resolutionHeight: str
    server: str
    port: str
    nativesDirectory: str
    enableLoggingConfig: bool
    disableMultiplayer: bool
    disableChat: bool
    quickPlayPath: str | None
    quickPlaySingleplayer: str | None
    quickPlayMultiplayer: str | None
    quickPlayRealms: str | None
    gameDir: str | None


class CallbackDict(TypedDict, total=False):
    setStatus: Callable[[str], None]
    setProgress: Callable[[int], None]
    setMax: Callable[[int], None]


class LatestMinecraftVersions(TypedDict):
    release: str
    snapshot: str


class MinecraftVersionInfo(TypedDict):
    id: str
    type: str
    releaseTime: datetime.datetime
    complianceLevel: int


# fabric


class FabricMinecraftVersion(TypedDict):
    version: str
    stable: bool


class FabricLoader(TypedDict):
    separator: str
    build: int
    maven: str
    version: str
    stable: bool


# quilt


class QuiltMinecraftVersion(TypedDict):
    version: str
    stable: bool


class QuiltLoader(TypedDict):
    separator: str
    build: int
    maven: str
    version: str


# java_utils


class JavaInformation(TypedDict):
    path: str
    name: str
    version: str
    java_path: str
    javaw_path: str | None
    is_64bit: bool
    openjdk: bool


# vanilla_launcher


class VanillaLauncherProfileResolution(TypedDict):
    height: int
    width: int


class VanillaLauncherProfile(TypedDict, total=False):
    name: str
    version: str | None
    versionType: Literal["latest-release", "latest-snapshot", "custom"]
    gameDirectory: str | None
    javaExecutable: str | None
    javaArguments: list[str] | None
    customResolution: VanillaLauncherProfileResolution | None


# mrpack


class MrpackInformation(TypedDict):
    name: str
    summary: str
    versionId: str
    formatVersion: int
    minecraftVersion: str
    optionalFiles: list[str]


class MrpackInstallOptions(TypedDict, total=False):
    optionalFiles: list[str]
    skipDependenciesInstall: bool


# runtime


class JvmRuntimeInformation(TypedDict):
    name: str
    released: datetime.datetime


class VersionRuntimeInformation(TypedDict):
    name: str
    javaMajorVersion: int


class _NewsEntryPlayPageImage(TypedDict):
    title: str
    url: str


class _NewsEntryNewsPageImageDimensions(TypedDict):
    width: int
    height: int


class _NewsEntryNewsPageImage(TypedDict):
    title: str
    url: str
    dimensions: _NewsEntryNewsPageImageDimensions


class _NewsEntry(TypedDict):
    title: str
    category: str
    date: str
    text: str
    playPageImage: _NewsEntryPlayPageImage
    newsPageImage: _NewsEntryNewsPageImage
    readMoreLink: str
    newsType: list[str]
    id: str


class MinecraftNews(TypedDict):
    version: Literal[1]
    entries: list[_NewsEntry]


class _JavaPatchNoteEntryImage(TypedDict):
    url: str
    title: str


class _JavaPatchNoteEntry(TypedDict):
    title: str
    type: Literal["release", "snapshot"]
    version: str
    image: _JavaPatchNoteEntryImage
    body: str
    contentPath: str


class JavaPatchNotes(TypedDict):
    version: Literal[1]
    entries: list[_JavaPatchNoteEntry]


@dataclass
class Credential:
    access_token: str
    username: str
    uuid: UUID


class Skin(TypedDict):
    skin: str
    cape: str


class MinecraftProfileProperty(TypedDict, total=False):
    name: str
    value: str
    signature: str


class MinecraftProfileSkin(TypedDict, total=False):
    id: str
    state: str
    url: str
    variant: str
    alias: str


class MinecraftProfileCape(TypedDict, total=False):
    id: str
    state: str
    url: str
    alias: str


class MinecraftProfileResponse(TypedDict, total=False):
    id: str
    name: str
    properties: list[MinecraftProfileProperty]  # 舊版
    skins: list[MinecraftProfileSkin]  # 新版
    capes: list[MinecraftProfileCape]  # 新版



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"runtime allows to install the java runtime. This module is used by :func:`~launcher_coreinstall.install_minecraft_version`, so you don't need to use it in your code most of the time."
from ._helper import (
    get_user_agent,
    download_file,
    empty,
    get_sha1_hash,
    check_path_inside_minecraft_directory,
    get_client_json,
)
from ._internal_types.runtime_types import (
    RuntimeListJson,
    PlatformManifestJson,
    _PlatformManifestJsonFile,
)
from ._types import CallbackDict, JvmRuntimeInformation, VersionRuntimeInformation
from .exceptions import VersionNotFound, PlatformNotSupported
import asyncio
import aiohttp
import aiofiles
import subprocess
import datetime
import platform
import os

_JVM_MANIFEST_URL = "https://launchermeta.mojang.com/v1/products/java-runtime/2ec0cc96c44e5a76b9c8b7c39df7210883d12871/all.json"


def _get_jvm_platform_string() -> str:
    """
    Get the name that is used the identify the platform
    """
    match platform.system():
        case "Windows":
            if platform.architecture()[0] == "32bit":
                return "windows-x86"
            else:
                return "windows-x64"
        case "Linux":
            if platform.architecture()[0] == "32bit":
                return "linux-i386"
            else:
                return "linux"
        case "Darwin":
            if platform.machine() == "arm64":
                return "mac-os-arm64"
            else:
                return "mac-os"
        case _:
            return "gamecore"


async def get_jvm_runtimes() -> list[str]:
    """
    Returns a list of all jvm runtimes

    Example:

    .. code:: python

        async for runtime in await launcher_coreruntime.get_jvm_runtimes():
            print(runtime)
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(
            _JVM_MANIFEST_URL, headers={"user-agent": get_user_agent()}
        ) as response:
            manifest_data: RuntimeListJson = await response.json()

    jvm_list = []
    for key in manifest_data[_get_jvm_platform_string()].keys():
        jvm_list.append(key)
    return jvm_list


async def get_installed_jvm_runtimes(
    minecraft_directory: str | os.PathLike,
) -> list[str]:
    """
    Returns a list of all installed jvm runtimes

    Example:

    .. code:: python

        for runtime in await launcher_coreruntime.get_installed_jvm_runtimes():
            print(runtime)

    :param minecraft_directory: The path to your Minecraft directory
    """
    try:
        return os.listdir(os.path.join(minecraft_directory, "runtime"))
    except FileNotFoundError:
        return []


async def install_jvm_runtime(
    jvm_version: str,
    minecraft_directory: str | os.PathLike,
    callback: CallbackDict | None = None,
    max_concurrency: int | None = None,
) -> None:
    """
    Installs the given jvm runtime. callback is the same dict as in the install module.

    Example:

    .. code:: python

        runtime_version = "java-runtime-gamma"
        minecraft_directory = launcher_coreutils.get_minecraft_directory()
        await launcher_coreruntime.install_jvm_runtime(runtime_version, minecraft_directory)

    :param jvm_version: The Name of the JVM version
    :param minecraft_directory: The path to your Minecraft directory
    :param callback: the same dict as for :func:`~launcher_coreinstall.install_minecraft_version`
    :param max_concurrency: number of concurrent tasks for asynchronous downloads. If None, it will be set automatically.
    :raises VersionNotFound: The given JVM Version was not found
    :raises FileOutsideMinecraftDirectory: A File should be placed outside the given Minecraft directory
    """
    if callback is None:
        callback = {}

    async with aiohttp.ClientSession() as session:
        async with session.get(
            _JVM_MANIFEST_URL, headers={"user-agent": get_user_agent()}
        ) as response:
            manifest_data: RuntimeListJson = await response.json()

        platform_string = _get_jvm_platform_string()
        # Check if the jvm version exists
        if jvm_version not in manifest_data[platform_string]:
            raise VersionNotFound(jvm_version)
        # Check if there is a platform manifest
        if len(manifest_data[platform_string][jvm_version]) == 0:
            return

        async with session.get(
            manifest_data[platform_string][jvm_version][0]["manifest"]["url"],
            headers={"user-agent": get_user_agent()},
        ) as response:
            platform_manifest: PlatformManifestJson = await response.json()

    base_path = os.path.join(
        minecraft_directory, "runtime", jvm_version, platform_string, jvm_version
    )
    file_list: list[str] = []

    async def install_runtime_file(key: str, value: _PlatformManifestJsonFile) -> None:
        """Install the single runtime file."""
        current_path = os.path.join(base_path, key)
        check_path_inside_minecraft_directory(minecraft_directory, current_path)

        if value["type"] == "file":
            # Prefer downloading the compresses file
            if "lzma" in value["downloads"]:
                await download_file(
                    value["downloads"]["lzma"]["url"],
                    current_path,
                    sha1=value["downloads"]["raw"]["sha1"],
                    callback=callback,
                    lzma_compressed=True,
                )
            else:
                await download_file(
                    value["downloads"]["raw"]["url"],
                    current_path,
                    sha1=value["downloads"]["raw"]["sha1"],
                    callback=callback,
                )

            # Make files executable on unix systems
            if value["executable"]:
                try:
                    subprocess.run(["chmod", "+x", current_path])
                except FileNotFoundError:
                    pass
            file_list.append(key)

        elif value["type"] == "directory":
            try:
                os.makedirs(current_path)
            except Exception:
                pass

        elif value["type"] == "link":
            check_path_inside_minecraft_directory(
                minecraft_directory, os.path.join(base_path, value["target"])
            )
            os.makedirs(os.path.dirname(current_path), exist_ok=True)

            try:
                os.symlink(value["target"], current_path)
            except Exception:
                pass

    # Download all files of the runtime
    callback.get("setMax", empty)(len(platform_manifest["files"]) - 1)
    count = 0
    sem = asyncio.Semaphore(max_concurrency if max_concurrency else 10)

    async def bounded_install(key: str, value: _PlatformManifestJsonFile) -> None:
        nonlocal count
        async with sem:
            await install_runtime_file(key, value)
            count += 1
            callback.get("setProgress", empty)(count)

    # Create and run tasks for all files
    tasks = [
        bounded_install(key, value) for key, value in platform_manifest["files"].items()
    ]
    await asyncio.gather(*tasks)

    # Create the .version file
    version_path = os.path.join(
        minecraft_directory, "runtime", jvm_version, platform_string, ".version"
    )
    check_path_inside_minecraft_directory(minecraft_directory, version_path)
    async with aiofiles.open(version_path, "w", encoding="utf-8") as f:
        await f.write(manifest_data[platform_string][jvm_version][0]["version"]["name"])

    # Writes the .sha1 file
    # It has the structure {path} /#// {sha1} {creation time in nanoseconds}
    sha1_path = os.path.join(
        minecraft_directory,
        "runtime",
        jvm_version,
        platform_string,
        f"{jvm_version}.sha1",
    )
    check_path_inside_minecraft_directory(minecraft_directory, sha1_path)
    async with aiofiles.open(sha1_path, "w", encoding="utf-8") as f:
        for current_file in file_list:
            current_path = os.path.join(base_path, current_file)
            ctime = os.stat(current_path).st_ctime_ns
            sha1 = get_sha1_hash(current_path)
            await f.write(f"{current_file} /#// {sha1} {ctime}\n")


async def get_executable_path(
    jvm_version: str, minecraft_directory: str | os.PathLike
) -> str | None:
    """
    Returns the path to the executable. Returns None if none is found.

    Example:

    .. code:: python

        runtime_version = "java-runtime-gamma"
        minecraft_directory = launcher_coreutils.get_minecraft_directory()
        executable_path = await launcher_coreruntime.get_executable_path(runtime_version, minecraft_directory)
        if executable_path is not None:
            print(f"Executable path: {executable_path}")
        else:
            print("The executable path was not found")

    :param jvm_version: The Name of the JVM version
    :param minecraft_directory: The path to your Minecraft directory
    """
    java_path = os.path.join(
        minecraft_directory,
        "runtime",
        jvm_version,
        _get_jvm_platform_string(),
        jvm_version,
        "bin",
        "java",
    )
    if os.path.isfile(java_path):
        return java_path
    elif os.path.isfile(java_path + ".exe"):
        return java_path + ".exe"
    java_path = java_path.replace(
        os.path.join("bin", "java"),
        os.path.join("jre.bundle", "Contents", "Home", "bin", "java"),
    )
    if os.path.isfile(java_path):
        return java_path
    else:
        return None


async def get_jvm_runtime_information(jvm_version: str) -> JvmRuntimeInformation:
    """
    Returns some Information about a JVM Version

    Example:

    .. code:: python

        runtime_version = "java-runtime-gamma"
        information = await launcher_coreruntime.get_jvm_runtime_information(runtime_version)
        print("Java version: " + information["name"])
        print("Release date: " + information["released"].isoformat())

    :param jvm_version: A JVM Version
    :raises VersionNotFound: The given JVM Version was not found
    :raises VersionNotFound: The given JVM Version is not available on this Platform
    :return: A Dict with Information
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(
            _JVM_MANIFEST_URL, headers={"user-agent": get_user_agent()}
        ) as response:
            manifest_data: RuntimeListJson = await response.json()

    platform_string = _get_jvm_platform_string()

    # Check if the jvm version exists
    if jvm_version not in manifest_data[platform_string]:
        raise VersionNotFound(jvm_version)

    if len(manifest_data[platform_string][jvm_version]) == 0:
        raise PlatformNotSupported()

    return {
        "name": manifest_data[platform_string][jvm_version][0]["version"]["name"],
        "released": datetime.datetime.fromisoformat(
            manifest_data[platform_string][jvm_version][0]["version"]["released"]
        ),
    }


async def get_version_runtime_information(
    version: str, minecraft_directory: str | os.PathLike
) -> VersionRuntimeInformation | None:
    """
    Returns information about the runtime used by a version

    Example:

    .. code:: python

        minecraft_version = "1.20"
        minecraft_directory = launcher_coreutils.get_minecraft_directory()
        information = await launcher_coreruntime.get_version_runtime_information(minecraft_version, minecraft_directory)
        print("Name: " + information["name"])
        print("Java version: " + str(information["javaMajorVersion"]))

    :param minecraft_directory: The path to your Minecraft directory
    :raises VersionNotFound: The Minecraft version was not found
    :return: A Dict with Information. None if the version has no runtime information.
    """
    data = await get_client_json(version, minecraft_directory)

    if "javaVersion" not in data:
        return None

    return {
        "name": data["javaVersion"]["component"],
        "javaMajorVersion": data["javaVersion"]["majorVersion"],
    }



# 標準庫導入
from typing import Literal
import platform
import zipfile
import json
import os

# 第三方庫導入
import aiofiles

# 本地導入
from ._internal_types.shared_types import ClientJson, ClientJsonLibrary
from ._helper import parse_rule_list, inherit_json, get_library_path
from .exceptions import VersionNotFound

__all__ = ["extract_natives"]


def get_natives(data: ClientJsonLibrary) -> str:
    """
    Returns the native part from the json data
    """
    if platform.architecture()[0] == "32bit":
        arch_type = "32"
    else:
        arch_type = "64"

    if "natives" in data:
        if platform.system() == "Windows":
            if "windows" in data["natives"]:
                return data["natives"]["windows"].replace("${arch}", arch_type)
            return ""
        if platform.system() == "Darwin":
            if "osx" in data["natives"]:
                return data["natives"]["osx"].replace("${arch}", arch_type)
            return ""
        if "linux" in data["natives"]:
            return data["natives"]["linux"].replace("${arch}", arch_type)
        return ""
    return ""


async def extract_natives_file(
    filename: str, extract_path: str, extract_data: dict[Literal["exclude"], list[str]]
) -> None:
    """
    Unpack natives
    """
    try:
        os.mkdir(extract_path)
    except (IOError, OSError):
        pass

    with zipfile.ZipFile(filename, "r") as zf:
        for i in zf.namelist():
            for e in extract_data["exclude"]:
                if i.startswith(e):
                    break
            else:
                zf.extract(i, extract_path)


async def extract_natives(
    versionid: str, path: str | os.PathLike, extract_path: str
) -> None:
    """
    Extract all native libraries from a version into the given directory. The directory will be created, if it does not exist.

    :param version: The Minecraft version
    :param minecraft_directory: The path to your Minecraft directory
    :param callback: The same dict as for :func:`~launcher_coreinstall.install_minecraft_version`
    :raises VersionNotFound: The Minecraft version was not found
    :raises FileOutsideMinecraftDirectory: A File should be placed outside the given Minecraft directory

    The natives are all extracted while installing. So you don't need to use this function in most cases.
    """
    if not os.path.isfile(
        os.path.join(path, "versions", versionid, versionid + ".json")
    ):
        raise VersionNotFound(versionid)

    async with aiofiles.open(
        os.path.join(path, "versions", versionid, versionid + ".json"),
        "r",
        encoding="utf-8",
    ) as f:
        data: ClientJson = json.loads(await f.read())

    if "inheritsFrom" in data:
        data = await inherit_json(data, path)

    for i in data["libraries"]:
        # Check, if the rules allow this lib for the current system
        if "rules" in i and not parse_rule_list(i["rules"], {}):
            continue

        current_path = get_library_path(i["name"], path)
        native = get_natives(i)

        if native == "":
            continue

        lib_path, extension = os.path.splitext(current_path)
        await extract_natives_file(
            f"{lib_path}-{native}{extension}",
            extract_path,
            i.get("extract", {"exclude": []}),
        )



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"""
.. note::
    Before using this module, please read this comment from the forge developers:

    .. code:: text

        Please do not automate the download and installation of Forge.
        Our efforts are supported by ads from the download page.
        If you MUST automate this, please consider supporting the project through https://www.patreon.com/LexManos/

    It's your choice, if you want to respect that and support forge.

forge contains functions for dealing with the Forge modloader
"""
from ._helper import (
    download_file,
    get_library_path,
    get_jar_mainclass,
    parse_maven_metadata,
    empty,
    extract_file_from_zip,
    get_classpath_separator,
    SUBPROCESS_STARTUP_INFO,
)
from .install import install_minecraft_version, install_libraries
from ._internal_types.forge_types import ForgeInstallProfile
from .exceptions import VersionNotFound
from ._types import CallbackDict
import asyncio
import aiofiles
import tempfile
import zipfile
import subprocess
import json
import os

__all__ = [
    "install_forge_version",
    "run_forge_installer",
    "list_forge_versions",
    "find_forge_version",
    "is_forge_version_valid",
    "supports_automatic_install",
    "forge_to_installed_version",
]


async def forge_processors(
    data: ForgeInstallProfile,
    minecraft_directory: str | os.PathLike,
    lzma_path: str,
    installer_path: str,
    callback: CallbackDict,
    java: str,
) -> None:
    """
    Run the processors of the install_profile.json
    """
    path = str(minecraft_directory)

    argument_vars = {
        "{MINECRAFT_JAR}": os.path.join(
            path, "versions", data["minecraft"], data["minecraft"] + ".jar"
        )
    }
    for data_key, data_value in data["data"].items():
        if data_value["client"].startswith("[") and data_value["client"].endswith("]"):
            argument_vars["{" + data_key + "}"] = get_library_path(
                data_value["client"][1:-1], path
            )
        else:
            argument_vars["{" + data_key + "}"] = data_value["client"]

    with tempfile.TemporaryDirectory() as root_path:
        argument_vars["{INSTALLER}"] = installer_path
        argument_vars["{BINPATCH}"] = lzma_path
        argument_vars["{ROOT}"] = root_path
        argument_vars["{SIDE}"] = "client"

        classpath_seperator = get_classpath_separator()

        callback.get("setMax", empty)(len(data["processors"]))

        for count, i in enumerate(data["processors"]):
            if "client" not in i.get("sides", ["client"]):
                # Skip server side only processors
                continue
            callback.get("setStatus", empty)("Running processor " + i["jar"])
            # Get the classpath
            classpath = ""
            for c in i["classpath"]:
                classpath = classpath + get_library_path(c, path) + classpath_seperator
            classpath = classpath + get_library_path(i["jar"], path)
            mainclass = get_jar_mainclass(get_library_path(i["jar"], path))
            command = [java, "-cp", classpath, mainclass]
            for c in i["args"]:
                var = argument_vars.get(c, c)
                if var.startswith("[") and var.endswith("]"):
                    command.append(get_library_path(var[1:-1], path))
                else:
                    command.append(var)
            for argument_key, argument_value in argument_vars.items():
                for pos in range(len(command)):
                    command[pos] = command[pos].replace(argument_key, argument_value)

            process = await asyncio.create_subprocess_exec(
                *command, startupinfo=SUBPROCESS_STARTUP_INFO
            )
            await process.wait()
            callback.get("setProgress", empty)(count)


async def install_forge_version(
    versionid: str,
    path: str | os.PathLike,
    callback: CallbackDict | None = None,
    java: str | os.PathLike | None = None,
) -> None:
    """
    Installs the given Forge version

    :param versionid: A Forge Version. You can get a List of Forge versions using :func:`list_forge_versions`
    :param path: The path to your Minecraft directory
    :param callback: The same dict as for :func:`~launcher_coreinstall.install_minecraft_version`
    :param java: A Path to a custom Java executable

    Raises a :class:`~launcher_coreexceptions.VersionNotFound` exception when the given forge version is not found
    """
    if callback is None:
        callback = {}

    FORGE_DOWNLOAD_URL = "https://maven.minecraftforge.net/net/minecraftforge/forge/{version}/forge-{version}-installer.jar"

    with tempfile.TemporaryDirectory(
        prefix="minecraft-launcher-lib-forge-install-"
    ) as tempdir:
        installer_path = os.path.join(tempdir, "installer.jar")

        if not await download_file(
            FORGE_DOWNLOAD_URL.format(version=versionid), installer_path, callback
        ):
            raise VersionNotFound(versionid)

        zf = zipfile.ZipFile(installer_path, "r")

        # Read the install_profile.json
        with zf.open("install_profile.json", "r") as f:
            version_content = f.read()

        version_data: ForgeInstallProfile = json.loads(version_content)
        forge_version_id = (
            version_data["version"]
            if "version" in version_data
            else version_data["install"]["version"]
        )
        minecraft_version = (
            version_data["minecraft"]
            if "minecraft" in version_data
            else version_data["install"]["minecraft"]
        )

        # Make sure, the base version is installed
        await install_minecraft_version(minecraft_version, path, callback=callback)

        # Install all needed libs from install_profile.json
        if "libraries" in version_data:
            await install_libraries(
                minecraft_version, version_data["libraries"], str(path), callback
            )

        # Extract the client.json
        version_json_path = os.path.join(
            path, "versions", forge_version_id, forge_version_id + ".json"
        )
        try:
            await extract_file_from_zip(
                zf, "version.json", version_json_path, minecraft_directory=path
            )
        except KeyError:
            if "versionInfo" in version_data:
                async with aiofiles.open(version_json_path, "w", encoding="utf-8") as f:
                    await f.write(
                        json.dumps(
                            version_data["versionInfo"], ensure_ascii=False, indent=4
                        )
                    )

        # Extract forge libs from the installer
        forge_lib_path = os.path.join(
            path, "libraries", "net", "minecraftforge", "forge", versionid
        )
        try:
            await extract_file_from_zip(
                zf,
                "maven/net/minecraftforge/forge/{version}/forge-{version}-universal.jar".format(
                    version=versionid
                ),
                os.path.join(forge_lib_path, "forge-" + versionid + "-universal.jar"),
                minecraft_directory=path,
            )
        except KeyError:
            pass

        try:
            await extract_file_from_zip(
                zf,
                "forge-{version}-universal.jar".format(version=versionid),
                os.path.join(forge_lib_path, f"forge-{versionid}.jar"),
                minecraft_directory=path,
            )
        except KeyError:
            pass

        try:
            await extract_file_from_zip(
                zf,
                f"maven/net/minecraftforge/forge/{versionid}/forge-{versionid}.jar",
                os.path.join(forge_lib_path, f"forge-{versionid}.jar"),
                minecraft_directory=path,
            )
        except KeyError:
            pass

        # Extract the client.lzma
        lzma_path = os.path.join(tempdir, "client.lzma")
        try:
            await extract_file_from_zip(zf, "data/client.lzma", lzma_path)
        except KeyError:
            pass

        zf.close()

        # Install the rest with the vanilla function
        await install_minecraft_version(forge_version_id, str(path), callback=callback)

        # Run the processors
        if "processors" in version_data:
            await forge_processors(
                version_data,
                str(path),
                lzma_path,
                installer_path,
                callback,
                "java" if java is None else str(java),
            )


async def run_forge_installer(
    version: str, java: str | os.PathLike | None = None
) -> None:
    """
    Run the forge installer of the given forge version

    :param version: A Forge Version. You can get a List of Forge versions using :func:`list_forge_versions`
    :param java: A Path to a custom Java executable
    """
    FORGE_DOWNLOAD_URL = "https://maven.minecraftforge.net/net/minecraftforge/forge/{version}/forge-{version}-installer.jar"

    with tempfile.TemporaryDirectory(
        prefix="minecraft-launcher-lib-forge-installer-"
    ) as tempdir:
        installer_path = os.path.join(tempdir, "installer.jar")

        if not await download_file(
            FORGE_DOWNLOAD_URL.format(version=version),
            installer_path,
            {},
            overwrite=True,
        ):
            raise VersionNotFound(version)

        process = await asyncio.create_subprocess_exec(
            "java" if java is None else str(java),
            "-jar",
            installer_path,
            cwd=tempdir,
            startupinfo=SUBPROCESS_STARTUP_INFO,
        )
        await process.wait()
        if process.returncode != 0:
            raise subprocess.CalledProcessError(
                process.returncode, ["java", "-jar", installer_path]
            )


async def list_forge_versions() -> list[str]:
    """
    Returns a list of all forge versions
    """
    MAVEN_METADATA_URL = (
        "https://maven.minecraftforge.net/net/minecraftforge/forge/maven-metadata.xml"
    )
    return (await parse_maven_metadata(MAVEN_METADATA_URL))["versions"]


async def find_forge_version(vanilla_version: str) -> str | None:
    """
    Find the latest forge version that is compatible to the given vanilla version

    :param vanilla_version: A vanilla Minecraft version
    """
    version_list = await list_forge_versions()
    for i in version_list:
        version_split = i.split("-")
        if version_split[0] == vanilla_version:
            return i
    return None


async def is_forge_version_valid(forge_version: str) -> bool:
    """
    Checks if a forge version is valid

    :param forge_version: A Forge Version
    """
    forge_version_list = await list_forge_versions()
    return forge_version in forge_version_list


async def supports_automatic_install(forge_version: str) -> bool:
    """
    Checks if install_forge_version() supports the given forge version

    :param forge_version: A Forge Version
    """
    try:
        vanilla_version, forge = forge_version.split("-")
        version_split = vanilla_version.split(".")
        version_number = int(version_split[1])
        if version_number >= 13:
            return True
        else:
            return False
    except Exception:
        return False


async def forge_to_installed_version(forge_version: str) -> str:
    """
    Returns the Version under which Forge will be installed from the given Forge version.

    :param forge_version: A Forge Version

    Raises a ValueError if the Version is invalid.
    """
    try:
        vanilla_part, forge_part = forge_version.split("-")
        return f"{vanilla_part}-forge-{forge_part}"
    except ValueError:
        raise ValueError(f"{forge_version} is not a valid forge version") from None



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"""
This module contains all Types for the :doc:`microsoft_account` module.
It has it's own module because of the many types needed that are not used somewhere else.
"""
from typing import Any, Literal, TypedDict


class AuthorizationTokenResponse(TypedDict):
    access_token: str
    token_type: Literal["Bearer"]
    expires_in: int
    scope: str
    refresh_token: str


class _Xui(TypedDict):
    uhs: str


class _DisplayClaims(TypedDict):
    xui: list[_Xui]


class XBLResponse(TypedDict):
    IssueInstant: str
    NotAfter: str
    Token: str
    DisplayClaims: _DisplayClaims


class XSTSResponse(TypedDict):
    IssueInstant: str
    NotAfter: str
    Token: str
    DisplayClaimns: _DisplayClaims


class _MinecraftStoreItem(TypedDict):
    name: str
    signature: str


class MinecraftStoreResponse(TypedDict):
    items: list[_MinecraftStoreItem]
    signature: str
    keyId: str


class MinecraftAuthenticateResponse(TypedDict):
    username: str
    roles: list[Any]
    access_token: str
    token_type: str
    expires_in: int


class _MinecraftProfileInfo(TypedDict):
    id: str
    state: Literal["ACTIVE", "INACTIVE"]
    url: str


class _MinecraftProfileSkin(_MinecraftProfileInfo):
    variant: str


class _MinecraftProfileCape(_MinecraftProfileInfo):
    alias: str


class MinecraftProfileResponse(TypedDict):
    id: str
    name: str
    skins: list[_MinecraftProfileSkin]
    capes: list[_MinecraftProfileCape]
    error: str
    errorMessage: str


class CompleteLoginResponse(MinecraftProfileResponse):
    access_token: str
    refresh_token: str



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"utils contains a few functions for helping you that doesn't fit in any other category"
from ._types import MinecraftOptions, LatestMinecraftVersions, MinecraftVersionInfo
from ._internal_types.shared_types import ClientJson, VersionListManifestJson
from ._helper import get_requests_response_cache, assert_func
from datetime import datetime
import platform
import pathlib
import random
import shutil
import uuid
import json
import os
import asyncio
import aiofiles


async def get_minecraft_directory() -> str:
    """
    Returns the default path to the .minecraft directory

    Example:

    .. code:: python

        minecraft_directory = await launcher_coreutils.get_minecraft_directory()
        print(f"The default minecraft directory is {minecraft_directory}")
    """
    if platform.system() == "Windows":
        return os.path.join(
            os.getenv(
                "APPDATA", os.path.join(pathlib.Path.home(), "AppData", "Roaming")
            ),
            ".minecraft",
        )
    elif platform.system() == "Darwin":
        return os.path.join(
            str(pathlib.Path.home()), "Library", "Application Support", "minecraft"
        )
    else:
        return os.path.join(str(pathlib.Path.home()), ".minecraft")


async def get_latest_version() -> LatestMinecraftVersions:
    """
    Returns the latest version of Minecraft

    Example:

    .. code:: python

        latest_version = await launcher_coreutils.get_latest_version()
        print("Latest Release " + latest_version["release"])
        print("Latest Snapshot " + latest_version["snapshot"])
    """
    response = await get_requests_response_cache(
        "https://launchermeta.mojang.com/mc/game/version_manifest_v2.json"
    )

    data = json.loads(response["content"])
    return data["latest"]


async def get_version_list() -> list[MinecraftVersionInfo]:
    """
    Returns all versions that Mojang offers to download

    Example:

    .. code:: python

        async for version in launcher_coreutils.get_version_list():
            print(version["id"])
    """
    response = await get_requests_response_cache(
        "https://launchermeta.mojang.com/mc/game/version_manifest_v2.json"
    )
    vlist: VersionListManifestJson = response
    returnlist: list[MinecraftVersionInfo] = []
    for i in vlist["versions"]:
        returnlist.append(
            {
                "id": i["id"],
                "type": i["type"],
                "releaseTime": datetime.fromisoformat(i["releaseTime"]),
                "complianceLevel": i["complianceLevel"],
            }
        )
    return returnlist


async def get_installed_versions(
    minecraft_directory: str | os.PathLike,
) -> list[MinecraftVersionInfo]:
    """
    Returns all installed versions

    Example:

    .. code:: python

        minecraft_directory = await launcher_coreutils.get_minecraft_directory()
        versions = await launcher_coreutils.get_installed_versions(minecraft_directory)
        for version in versions:
            print(version["id"])

    :param minecraft_directory: The path to your Minecraft directory
    """
    try:
        dir_list = await asyncio.to_thread(
            os.listdir, os.path.join(minecraft_directory, "versions")
        )
    except FileNotFoundError:
        return []

    version_list: list[MinecraftVersionInfo] = []
    for i in dir_list:
        json_path = os.path.join(minecraft_directory, "versions", i, i + ".json")
        if not await asyncio.to_thread(os.path.isfile, json_path):
            continue

        try:
            async with aiofiles.open(json_path, "r", encoding="utf-8") as f:
                version_data: ClientJson = json.loads(await f.read())

            try:
                release_time = datetime.fromisoformat(version_data["releaseTime"])
            except ValueError:
                # In case some custom client has a invalid time
                release_time = datetime.fromtimestamp(0)

            version_list.append(
                {
                    "id": version_data["id"],
                    "type": version_data["type"],
                    "releaseTime": release_time,
                    "complianceLevel": version_data.get("complianceLevel", 0),
                }
            )
        except Exception:
            continue

    return version_list


async def get_available_versions(
    minecraft_directory: str | os.PathLike,
) -> list[MinecraftVersionInfo]:
    """
    Returns all installed versions and all versions that Mojang offers to download

    Example:

    .. code:: python

        minecraft_directory = await launcher_coreutils.get_minecraft_directory()
        versions = await launcher_coreutils.get_available_versions(minecraft_directory)
        for version in versions:
            print(version["id"])

    :param minecraft_directory: The path to your Minecraft directory
    """
    version_list = []
    version_check = []

    for i in await get_version_list():
        version_list.append(i)
        version_check.append(i["id"])

    for i in await get_installed_versions(minecraft_directory):
        if i["id"] not in version_check:
            version_list.append(i)

    return version_list


async def get_java_executable() -> str:
    """
    Tries the find out the path to the default java executable.
    Returns :code:`java`, if no path was found.

    Example:

    .. code:: python

        print("The path to Java is " + await launcher_coreutils.get_java_executable())
    """
    if platform.system() == "Windows":
        if (java_home := os.getenv("JAVA_HOME")) is not None:
            return os.path.join(java_home, "bin", "javaw.exe")
        elif await asyncio.to_thread(
            os.path.isfile,
            r"C:\Program Files (x86)\Common Files\Oracle\Java\javapath\javaw.exe",
        ):
            return r"C:\Program Files (x86)\Common Files\Oracle\Java\javapath\javaw.exe"
        else:
            return await asyncio.to_thread(shutil.which, "javaw") or "javaw"
    elif (java_home := os.getenv("JAVA_HOME")) is not None:
        return os.path.join(java_home, "bin", "java")
    elif platform.system() == "Darwin":
        return await asyncio.to_thread(shutil.which, "java") or "java"
    else:
        if await asyncio.to_thread(os.path.islink, "/etc/alternatives/java"):
            return await asyncio.to_thread(os.readlink, "/etc/alternatives/java")
        elif await asyncio.to_thread(os.path.islink, "/usr/lib/jvm/default-runtime"):
            return os.path.join(
                "/usr",
                "lib",
                "jvm",
                await asyncio.to_thread(os.readlink, "/usr/lib/jvm/default-runtime"),
                "bin",
                "java",
            )
        else:
            return await asyncio.to_thread(shutil.which, "java") or "java"


_version_cache = None


async def get_library_version() -> str:
    """
    Returns the version of minecraft-launcher-lib

    Example:

    .. code:: python

        print(f"You are using version {await launcher_coreutils.get_library_version()} of minecraft-launcher-lib")
    """
    global _version_cache
    if _version_cache is not None:
        return _version_cache
    else:
        version_path = os.path.join(os.path.dirname(__file__), "version.txt")
        async with aiofiles.open(version_path, "r", encoding="utf-8") as f:
            _version_cache = (await f.read()).strip()
            return _version_cache


async def generate_test_options() -> MinecraftOptions:
    """
    Generates test options to launch minecraft.
    This includes a random name and a random uuid.

    .. note::
        This function is just for debugging and testing, if Minecraft works.
        The behavior of this function may change in the future.
        Do not use it in production.

    Example:

    .. code:: python

        version = "1.0"
        options = await launcher_coreutils.generate_test_options()
        minecraft_directory = await launcher_coreutils.get_minecraft_directory()
        command = await launcher_corecommand.get_minecraft_command(version, minecraft_directory, options)
        await asyncio.create_subprocess_exec(*command)
    """
    return {
        "username": f"Player{random.randrange(100, 1000)}",
        "uuid": str(uuid.uuid4()),
        "token": "",
    }


async def is_version_valid(
    version: str, minecraft_directory: str | os.PathLike
) -> bool:
    """
    Checks if the given version exists.
    This checks if the given version is installed or offered to download by Mojang.
    Basically you can use this tho check, if the given version can be used with :func:`~launcher_coreinstall.install_minecraft_version`.

    Example:

    .. code:: python

        version = "1.0"
        minecraft_directory = await launcher_coreutils.get_minecraft_directory()
        if await launcher_coreutils.is_version_valid(version, minecraft_directory):
            print(f"{version} is a valid version")
        else:
            print(f"{version} is not a valid version")

    :param version: A Minecraft version
    :param minecraft_directory: The path to your Minecraft directory
    """
    if await asyncio.to_thread(
        os.path.isdir, os.path.join(minecraft_directory, "versions", version)
    ):
        return True
    for i in await get_version_list():
        if i["id"] == version:
            return True
    return False


async def is_vanilla_version(version: str) -> bool:
    """
    Checks if the given version is a vanilla version

    Example:

    .. code:: python

        version = "1.0"
        if await launcher_coreutils.is_vanilla_version(version):
            print(f"{version} is a vanilla version")
        else:
            print(f"{version} is not a vanilla version")

    :param version: A Minecraft version
    """
    for i in await get_version_list():
        if i["id"] == version:
            return True
    return False


async def is_platform_supported() -> bool:
    """
    Checks if the current platform is supported

    Example:

    .. code:: python

        if not await launcher_coreutils.is_platform_supported():
            print("Your platform is not supported", file=sys.stderr)
            sys.exit(1)
    """
    if platform.system() not in ["Windows", "Darwin", "Linux"]:
        return False
    return True


async def is_minecraft_installed(minecraft_directory: str | os.PathLike) -> bool:
    """
    Checks, if there is already a existing Minecraft Installation in the given Directory

    Example:

    .. code:: python

        minecraft_directory = await launcher_coreutils.get_minecraft_directory()
        if await launcher_coreutils.is_minecraft_installed(minecraft_directory):
            print("Minecraft is installed")
        else:
            print("Minecraft is not installed")

    :param minecraft_directory: The path to your Minecraft directory
    :return: Is a Installation is found
    """
    try:
        versions_path = os.path.join(minecraft_directory, "versions")
        libraries_path = os.path.join(minecraft_directory, "libraries")
        assets_path = os.path.join(minecraft_directory, "assets")

        is_versions = await asyncio.to_thread(os.path.isdir, versions_path)
        is_libraries = await asyncio.to_thread(os.path.isdir, libraries_path)
        is_assets = await asyncio.to_thread(os.path.isdir, assets_path)

        await asyncio.to_thread(assert_func, is_versions)
        await asyncio.to_thread(assert_func, is_libraries)
        await asyncio.to_thread(assert_func, is_assets)
        return True
    except AssertionError:
        return False



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
from typing import Literal, TypedDict


class VanillaLauncherProfilesJsonProfile(TypedDict, total=False):
    created: str
    gameDir: str
    icon: str
    javaArgs: str
    javaDir: str
    lastUsed: str
    lastVersionId: str
    name: str
    resolution: dict[Literal["height", "width"], int]
    type: str


class VanillaLauncherProfilesJson(TypedDict):
    profiles: dict[str, VanillaLauncherProfilesJsonProfile]
    version: int



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"install allows you to install minecraft."
from ._helper import (
    download_file,
    parse_rule_list,
    inherit_json,
    empty,
    get_user_agent,
    check_path_inside_minecraft_directory,
)
from ._internal_types.shared_types import ClientJson, ClientJsonLibrary
from .natives import extract_natives_file, get_natives
from ._internal_types.install_types import AssetsJson
from .runtime import install_jvm_runtime
from .exceptions import VersionNotFound
from ._types import CallbackDict
import aiohttp
import aiofiles
import asyncio
import shutil
import json
import os

__all__ = ["install_minecraft_version"]


async def install_libraries(
    id: str,
    libraries: list[ClientJsonLibrary],
    path: str,
    callback: CallbackDict,
    max_workers: int | None = None,
) -> None:
    """
    Install all libraries
    """
    callback.get("setStatus", empty)("Download Libraries")
    callback.get("setMax", empty)(len(libraries) - 1)

    async def download_library(
        i: ClientJsonLibrary,
    ) -> None:
        """Download the single library."""
        # Check, if the rules allow this lib for the current system
        if "rules" in i and not parse_rule_list(i["rules"], {}):
            return

        # Turn the name into a path
        current_path = os.path.join(path, "libraries")
        if "url" in i:
            if i["url"].endswith("/"):
                download_url = i["url"][:-1]
            else:
                download_url = i["url"]
        else:
            download_url = "https://libraries.minecraft.net"

        try:
            lib_path, name, version = i["name"].split(":")[0:3]
        except ValueError:
            return

        for lib_part in lib_path.split("."):
            current_path = os.path.join(current_path, lib_part)
            download_url = f"{download_url}/{lib_part}"

        try:
            version, fileend = version.split("@")
        except ValueError:
            fileend = "jar"

        jar_filename = f"{name}-{version}.{fileend}"
        download_url = f"{download_url}/{name}/{version}"
        current_path = os.path.join(current_path, name, version)
        native = get_natives(i)

        # Check if there is a native file
        if native != "":
            jar_filename_native = f"{name}-{version}-{native}.jar"
        jar_filename = f"{name}-{version}.{fileend}"
        download_url = f"{download_url}/{jar_filename}"

        # Try to download the lib
        try:
            async with aiohttp.ClientSession() as session:
                await download_file(
                    download_url,
                    os.path.join(current_path, jar_filename),
                    callback=callback,
                    session=session,
                    minecraft_directory=path,
                )
        except Exception:
            pass

        if "downloads" not in i:
            if "extract" in i:
                await extract_natives_file(
                    os.path.join(current_path, jar_filename_native),
                    os.path.join(path, "versions", id, "natives"),
                    i["extract"],
                )
            return

        if (
            "artifact" in i["downloads"]
            and i["downloads"]["artifact"]["url"] != ""
            and "path" in i["downloads"]["artifact"]
        ):
            async with aiohttp.ClientSession() as session:
                await download_file(
                    i["downloads"]["artifact"]["url"],
                    os.path.join(path, "libraries", i["downloads"]["artifact"]["path"]),
                    callback,
                    sha1=i["downloads"]["artifact"]["sha1"],
                    session=session,
                    minecraft_directory=path,
                )
        if native != "":
            async with aiohttp.ClientSession() as session:
                await download_file(
                    i["downloads"]["classifiers"][native]["url"],
                    os.path.join(current_path, jar_filename_native),
                    callback,
                    sha1=i["downloads"]["classifiers"][native]["sha1"],
                    session=session,
                    minecraft_directory=path,
                )
            await extract_natives_file(
                os.path.join(current_path, jar_filename_native),
                os.path.join(path, "versions", id, "natives"),
                i.get("extract", {"exclude": []}),
            )

    count = 0
    tasks = []
    for i in libraries:
        task = asyncio.create_task(download_library(i))
        tasks.append(task)

    # Use semaphore to limit concurrent downloads if max_workers is specified
    if max_workers:
        semaphore = asyncio.Semaphore(max_workers)

        async def limited_download(lib):
            async with semaphore:
                await download_library(lib)

        tasks = [asyncio.create_task(limited_download(i)) for i in libraries]

    for task in asyncio.as_completed(tasks):
        await task
        count += 1
        callback.get("setProgress", empty)(count)


async def install_assets(
    data: ClientJson,
    path: str,
    callback: CallbackDict,
    max_workers: int | None = None,
) -> None:
    """
    Install all assets
    """
    # Old versions don't have this
    if "assetIndex" not in data:
        return

    callback.get("setStatus", empty)("Download Assets")

    # Download all assets
    async with aiohttp.ClientSession() as session:
        await download_file(
            data["assetIndex"]["url"],
            os.path.join(path, "assets", "indexes", data["assets"] + ".json"),
            callback,
            sha1=data["assetIndex"]["sha1"],
            session=session,
        )

    async with aiofiles.open(
        os.path.join(path, "assets", "indexes", data["assets"] + ".json"), "r"
    ) as f:
        assets_data: AssetsJson = json.loads(await f.read())

    # The assets has a hash. e.g. c4dbabc820f04ba685694c63359429b22e3a62b5
    # With this hash, it can be download from https://resources.download.minecraft.net/c4/c4dbabc820f04ba685694c63359429b22e3a62b5
    # And saved at assets/objects/c4/c4dbabc820f04ba685694c63359429b22e3a62b5
    assets = set(val["hash"] for val in assets_data["objects"].values())
    callback.get("setMax", empty)(len(assets) - 1)
    count = 0

    async def download_asset(filehash: str) -> None:
        """Download the single asset file."""
        async with aiohttp.ClientSession() as session:
            await download_file(
                "https://resources.download.minecraft.net/"
                + filehash[:2]
                + "/"
                + filehash,
                os.path.join(path, "assets", "objects", filehash[:2], filehash),
                callback,
                sha1=filehash,
                session=session,
                minecraft_directory=path,
            )

    # Use semaphore to limit concurrency if max_workers is specified
    if max_workers:
        semaphore = asyncio.Semaphore(max_workers)

        async def limited_download(filehash):
            async with semaphore:
                await download_asset(filehash)

        tasks = [asyncio.create_task(limited_download(filehash)) for filehash in assets]
    else:
        tasks = [asyncio.create_task(download_asset(filehash)) for filehash in assets]

    for task in asyncio.as_completed(tasks):
        await task
        count += 1
        callback.get("setProgress", empty)(count)


async def do_version_install(
    versionid: str,
    path: str,
    callback: CallbackDict,
    url: str | None = None,
    sha1: str | None = None,
) -> None:
    """
    Installs the given version
    """
    # Download and read versions.json
    if url:
        async with aiohttp.ClientSession() as session:
            await download_file(
                url,
                os.path.join(path, "versions", versionid, versionid + ".json"),
                callback,
                sha1=sha1,
                session=session,
                minecraft_directory=path,
            )

    async with aiofiles.open(
        os.path.join(path, "versions", versionid, versionid + ".json"),
        "r",
        encoding="utf-8",
    ) as f:
        versiondata: ClientJson = json.loads(await f.read())

    # For Forge
    if "inheritsFrom" in versiondata:
        try:
            await install_minecraft_version(
                versiondata["inheritsFrom"], path, callback=callback
            )
        except VersionNotFound:
            pass
        versiondata = await inherit_json(versiondata, path)

    await install_libraries(versiondata["id"], versiondata["libraries"], path, callback)
    await install_assets(versiondata, path, callback)

    # Download logging config
    if "logging" in versiondata:
        if len(versiondata["logging"]) != 0:
            logger_file = os.path.join(
                path,
                "assets",
                "log_configs",
                versiondata["logging"]["client"]["file"]["id"],
            )
            async with aiohttp.ClientSession() as session:
                await download_file(
                    versiondata["logging"]["client"]["file"]["url"],
                    logger_file,
                    callback,
                    sha1=versiondata["logging"]["client"]["file"]["sha1"],
                    session=session,
                    minecraft_directory=path,
                )

    # Download minecraft.jar
    if "downloads" in versiondata:
        async with aiohttp.ClientSession() as session:
            await download_file(
                versiondata["downloads"]["client"]["url"],
                os.path.join(
                    path, "versions", versiondata["id"], versiondata["id"] + ".jar"
                ),
                callback,
                sha1=versiondata["downloads"]["client"]["sha1"],
                session=session,
                minecraft_directory=path,
            )

    # Need to copy jar for old forge versions
    if (
        not os.path.isfile(
            os.path.join(
                path, "versions", versiondata["id"], versiondata["id"] + ".jar"
            )
        )
        and "inheritsFrom" in versiondata
    ):
        inherits_from = versiondata["inheritsFrom"]
        inherit_path = os.path.join(
            path, "versions", inherits_from, f"{inherits_from}.jar"
        )
        check_path_inside_minecraft_directory(path, inherit_path)
        shutil.copyfile(
            os.path.join(
                path, "versions", versiondata["id"], versiondata["id"] + ".jar"
            ),
            inherit_path,
        )

    # Install java runtime if needed
    if "javaVersion" in versiondata:
        callback.get("setStatus", empty)("Install java runtime")
        await install_jvm_runtime(
            versiondata["javaVersion"]["component"], path, callback=callback
        )

    callback.get("setStatus", empty)("Installation complete")


async def install_minecraft_version(
    versionid: str,
    minecraft_directory: str | os.PathLike,
    callback: CallbackDict | None = None,
) -> None:
    """
    Installs a minecraft version into the given path. e.g. ``install_version("1.14", "/tmp/minecraft")``. Use :func:`~launcher_coreutils.get_minecraft_directory` to get the default Minecraft directory.

    :param versionid: The Minecraft version
    :param minecraft_directory: The path to your Minecraft directory
    :param callback: Some functions that are called to monitor the progress (see below)
    :raises VersionNotFound: The Minecraft version was not found
    :raises FileOutsideMinecraftDirectory: A File should be placed outside the given Minecraft directory

    ``callback`` is a dict with functions that are called with arguments to get the progress. You can use it to show the progress to the user.

    .. code:: python

        callback = {
            "setStatus": some_function, # This function is called to set a text
            "setProgress" some_function, # This function is called to set the progress.
            "setMax": some_function, # This function is called to set to max progress.
        }

    Files that are already exists will not be replaced.
    """
    if isinstance(minecraft_directory, os.PathLike):
        minecraft_directory = str(minecraft_directory)
    if callback is None:
        callback = {}
    if os.path.isfile(
        os.path.join(minecraft_directory, "versions", versionid, f"{versionid}.json")
    ):
        await do_version_install(versionid, minecraft_directory, callback)
        return

    async with aiohttp.ClientSession(
        headers={"user-agent": get_user_agent()}
    ) as session:
        async with session.get(
            "https://launchermeta.mojang.com/mc/game/version_manifest_v2.json"
        ) as response:
            version_list = await response.json()

    for i in version_list["versions"]:
        if i["id"] == versionid:
            await do_version_install(
                versionid, minecraft_directory, callback, url=i["url"], sha1=i["sha1"]
            )
            return
    raise VersionNotFound(versionid)



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"exceptions contains all custom exceptions that can be raised by minecraft_launcher_lib"
from ._types import VanillaLauncherProfile


class VersionNotFound(ValueError):
    """
    The given version does not exists
    """

    def __init__(self, version: str) -> None:
        self.version: str = version
        "The version that caused the exception"

        self.msg: str = f"Version {version} was not found"
        "A message to display"

        ValueError.__init__(self, self.msg)


class UnsupportedVersion(ValueError):
    """
    This Exception is raised when you try to run :func:`~launcher_corefabric.install_fabric` or :func:`~launcher_corequilt.install_quilt` with a unsupported version
    """

    def __init__(self, version: str) -> None:
        self.version: str = version
        "The version that caused the exception"

        self.msg: str = f"Version {version} is not supported"
        "A message to display"

        ValueError.__init__(self, self.msg)


class ExternalProgramError(Exception):
    """
    This Exception is raised when a external program failed
    """

    def __init__(self, command: list[str], stdout: bytes, stderr: bytes) -> None:
        self.command: list[str] = command
        "The command that caused the error"

        self.stdout: bytes = stdout
        "The stdout of the command"

        self.stderr: bytes = stderr
        "The stderr of the command"


class InvalidRefreshToken(ValueError):
    """
    Raised when :func:`~launcher_coremicrosoft_account.complete_refresh` is called with a invalid refresh token
    """

    pass


class InvalidVanillaLauncherProfile(ValueError):
    """
    Raised when a function from the :doc:`vanilla_launcher` module is called with a invalid vanilla profile
    """

    def __init__(self, profile: VanillaLauncherProfile) -> None:
        self.profile: VanillaLauncherProfile = profile
        "The invalid profile"

        super().__init__("Invalid VanillaLauncherProfile")


class SecurityError(Exception):
    """
    Raised when something security related happens
    """

    def __init__(self, code: str, message: str) -> None:
        self.code: str = code
        "A Code to specify the Error"

        self.message: str = message
        "A Message to display"

        super().__init__(message)


class FileOutsideMinecraftDirectory(SecurityError):
    """
    Raised when a File should be placed outside the given Minecraft directory
    """

    def __init__(self, path: str, minecraft_directory: str) -> None:
        self.path: str = path
        "The Path of the File"

        self.minecraft_directory: str = minecraft_directory
        "The Minecraft directory of the File"

        super().__init__(
            "FileOutsideMinecraftDirectory",
            f"Tried to place {path} outside {minecraft_directory}",
        )


class InvalidChecksum(SecurityError):
    """
    Raised when a File did not match the Checksum
    """

    def __init__(
        self, url: str, path: str, expected_checksum: str, actual_checksum: str
    ) -> None:
        self.url: str = url
        "The URL to the File with the wrong Checksum"

        self.path: str = path
        "The Path to the File with the wrong Checksum"

        self.expected_checksum: str = expected_checksum
        "The expected Checksum"

        self.actual_checksum: str = actual_checksum
        "The actual Checksum"

        super().__init__(
            "InvalidChecksum",
            f"{path} has the wrong Checksum (expected {expected_checksum} got {actual_checksum})",
        )


class AzureAppNotPermitted(Exception):
    """
    Raised when you try to use a Azure App, that don't have the Permission to use the Minecraft API.
    Take a look at the For more information about the options take a look at the :doc:`/tutorial/microsoft_login` tutorial to learn how to fix this.
    """

    def __init__(self) -> None:
        super().__init__(
            "It looks like your Azure App don't have the Permission to use the Minecraft API. Take a look at the Documentation for more Information."
        )


class PlatformNotSupported(Exception):
    """
    Raised, when the current Platform is not supported by a feature
    """

    def __init__(self) -> None:
        super().__init__("Your Platform is not supported")


class AccountNotOwnMinecraft(Exception):
    """
    Raised by :func:`~launcher_coremicrosoft_account.complete_login` and :func:`~launcher_coremicrosoft_account.complete_login` when the Account does not own Minecraft
    """

    def __init__(self) -> None:
        super().__init__("This Account does not own Minecraft")


class AccountBanFromXbox(Exception):
    """
    Raised when the Account is banned from Xbox
    """

    def __init__(self) -> None:
        super().__init__("This Account is banned from Xbox")


class AccountNotHaveXbox(Exception):
    """
    Raised when the Account does not have Xbox
    """

    def __init__(self) -> None:
        super().__init__("This Account does not have Xbox")


class XboxLiveNotAvailable(Exception):
    """
    Raised when the Xbox Live Service is not available
    """

    def __init__(self) -> None:
        super().__init__("Xbox Live Service is not available")


class AccountNeedAdultVerification(Exception):
    """
    Raised when the Account needs adult verification
    """

    def __init__(self) -> None:
        super().__init__("This Account needs adult verification")



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"command contains the function for creating the minecraft command"
from ._helper import (
    parse_rule_list,
    inherit_json,
    get_classpath_separator,
    get_library_path,
)
from ._internal_types.shared_types import ClientJson, ClientJsonArgumentRule
from .runtime import get_executable_path
from .exceptions import VersionNotFound
from .utils import get_library_version
from ._types import MinecraftOptions, Credential
from .natives import get_natives
import aiofiles
import json
import copy
import os

__all__ = ["get_minecraft_command"]


def get_libraries(data: ClientJson, path: str) -> str:
    """
    Returns the argument with all libs that come after -cp
    """
    classpath_seperator = get_classpath_separator()
    libstr = ""
    for i in data["libraries"]:
        if "rules" in i and not parse_rule_list(i["rules"], {}):
            continue

        libstr += get_library_path(i["name"], path) + classpath_seperator
        native = get_natives(i)
        if native != "":
            if "downloads" in i and "path" in i["downloads"]["classifiers"][native]:  # type: ignore
                libstr += os.path.join(path, "libraries", i["downloads"]["classifiers"][native]["path"]) + classpath_seperator  # type: ignore
            else:
                libstr += (
                    get_library_path(i["name"] + "-" + native, path)
                    + classpath_seperator
                )

    if "jar" in data:
        libstr = libstr + os.path.join(
            path, "versions", data["jar"], data["jar"] + ".jar"
        )
    else:
        libstr = libstr + os.path.join(
            path, "versions", data["id"], data["id"] + ".jar"
        )

    return libstr


async def replace_arguments(
    argstr: str,
    versionData: ClientJson,
    path: str,
    options: MinecraftOptions,
    classpath: str,
) -> str:
    arg_replacements = {
        "${natives_directory}": options["nativesDirectory"],
        "${launcher_name}": options.get("launcherName", "minecraft-launcher-lib"),
        "${launcher_version}": options.get(
            "launcherVersion", await get_library_version()
        ),
        "${classpath}": classpath,
        "${auth_player_name}": options.get("username", "{username}"),
        "${version_name}": versionData["id"],
        "${game_directory}": options.get("gameDir", path),
        "${assets_root}": os.path.join(path, "assets"),
        "${assets_index_name}": versionData.get("assets", versionData["id"]),
        "${auth_uuid}": options.get("uuid", "{uuid}"),
        "${auth_access_token}": options.get("token", "{token}"),
        "${user_type}": "msa",
        "${version_type}": versionData["type"],
        "${user_properties}": "{}",
        "${resolution_width}": options.get("resolutionWidth", "854"),
        "${resolution_height}": options.get("resolutionHeight", "480"),
        "${game_assets}": os.path.join(path, "assets", "virtual", "legacy"),
        "${auth_session}": options.get("token", "{token}"),
        "${library_directory}": os.path.join(path, "libraries"),
        "${classpath_separator}": get_classpath_separator(),
        "${quickPlayPath}": options.get("quickPlayPath") or "{quickPlayPath}",
        "${quickPlaySingleplayer}": options.get("quickPlaySingleplayer")
        or "{quickPlaySingleplayer}",
        "${quickPlayMultiplayer}": options.get("quickPlayMultiplayer")
        or "{quickPlayMultiplayer}",
        "${quickPlayRealms}": options.get("quickPlayRealms") or "{quickPlayRealms}",
    }

    for key, value in arg_replacements.items():
        argstr = argstr.replace(key, str(value))

    return argstr


async def get_arguments_string(
    versionData: ClientJson, path: str, options: MinecraftOptions, classpath: str
) -> list[str]:
    """
    Turns the argument string from the client.json into a list
    """
    arglist: list[str] = []

    for v in versionData["minecraftArguments"].split(" "):
        v = await replace_arguments(v, versionData, path, options, classpath)
        arglist.append(v)

    # Custom resolution is not in the list
    if options.get("customResolution", False):
        arglist.append("--width")
        arglist.append(options.get("resolutionWidth", "854"))
        arglist.append("--height")
        arglist.append(options.get("resolutionHeight", "480"))

    if options.get("demo", False):
        arglist.append("--demo")

    return arglist


async def get_arguments(
    data: list[str | ClientJsonArgumentRule],
    versionData: ClientJson,
    path: str,
    options: MinecraftOptions,
    classpath: str,
) -> list[str]:
    """
    Returns all arguments from the client.json
    """
    arglist: list[str] = []
    for i in data:
        # i could be the argument
        if isinstance(i, str):
            arglist.append(
                await replace_arguments(i, versionData, path, options, classpath)
            )
        else:
            # Rules might has 2 different names in different client.json
            if "compatibilityRules" in i and not parse_rule_list(
                i["compatibilityRules"], options
            ):
                continue

            if "rules" in i and not parse_rule_list(i["rules"], options):
                continue

            # Sometimes  i["value"] is the argument
            if isinstance(i["value"], str):
                arglist.append(
                    await replace_arguments(
                        i["value"], versionData, path, options, classpath
                    )
                )
            # Sometimes i["value"] is a list of arguments
            else:
                for v in i["value"]:
                    v = await replace_arguments(
                        v, versionData, path, options, classpath
                    )
                    arglist.append(v)
    return arglist


async def get_minecraft_command(
    version: str,
    minecraft_directory: str | os.PathLike,
    options: MinecraftOptions,
    Credential: Credential | None = None,
) -> list[str]:
    """
    Returns the command for running minecraft as list. The given command can be executed with subprocess. Use :func:`~launcher_coreutils.get_minecraft_directory` to get the default Minecraft directory.

    :param version: The Minecraft version
    :param minecraft_directory: The path to your Minecraft directory
    :param options: Some Options (see below)

    ``options`` is a dict:

    .. code:: python

        options = {
            # This is needed
            "username": The Username,
            "uuid": uuid of the user,
            "token": the accessToken,
            # This is optional
            "executablePath": "java", # The path to the java executable
            "defaultExecutablePath": "java", # The path to the java executable if the client.json has none
            "jvmArguments": [], #The jvmArguments
            "launcherName": "minecraft-launcher-lib", # The name of your launcher
            "launcherVersion": "1.0", # The version of your launcher
            "gameDirectory": "/home/user/.minecraft", # The gameDirectory (default is the path given in arguments)
            "demo": False, # Run Minecraft in demo mode
            "customResolution": False, # Enable custom resolution
            "resolutionWidth": "854", # The resolution width
            "resolutionHeight": "480", # The resolution height
            "server": "example.com", # The IP of a server where Minecraft connect to after start
            "port": "123", # The port of a server where Minecraft connect to after start
            "nativesDirectory": "minecraft_directory/versions/version/natives", # The natives directory
            "enableLoggingConfig": False, # Enable use of the log4j configuration file
            "disableMultiplayer": False, # Disables the multiplayer
            "disableChat": False, # Disables the chat
            "quickPlayPath": None, # The Quick Play Path
            "quickPlaySingleplayer": None, # The Quick Play Singleplayer
            "quickPlayMultiplayer": None, # The Quick Play Multiplayer
            "quickPlayRealms": None, # The Quick Play Realms
        }

    You can use the :doc:`microsoft_account` module to get the needed information.
    For more information about the options take a look at the :doc:`/tutorial/more_launch_options` tutorial.
    """
    path = str(minecraft_directory)

    if Credential:
        token = Credential.access_token if Credential else options.get("token", None)
        username = Credential.username if Credential else options.get("username", None)
        uuid = Credential.uuid if Credential else options.get("uuid", None)
        options["token"] = token
        options["username"] = username
        options["uuid"] = uuid

    if not os.path.isdir(os.path.join(path, "versions", version)):
        raise VersionNotFound(version)

    options = copy.deepcopy(options)

    json_path = os.path.join(path, "versions", version, version + ".json")
    async with aiofiles.open(json_path, "r", encoding="utf-8") as f:
        data: ClientJson = json.loads(await f.read())

    if "inheritsFrom" in data:
        data = await inherit_json(data, path)

    options["nativesDirectory"] = options.get(
        "nativesDirectory", os.path.join(path, "versions", data["id"], "natives")
    )
    classpath = get_libraries(data, path)

    command: list[str] = []
    # Add Java executable
    if "executablePath" in options:
        command.append(options["executablePath"])
    elif "javaVersion" in data:
        java_path = await get_executable_path(data["javaVersion"]["component"], path)
        if java_path is None:
            command.append("java")
        else:
            command.append(java_path)
    else:
        command.append(options.get("defaultExecutablePath", "java"))

    if "jvmArguments" in options:
        command = command + options["jvmArguments"]

    # Newer Versions have jvmArguments in client.json
    if isinstance(data.get("arguments", None), dict):
        if "jvm" in data["arguments"]:
            command = command + await get_arguments(
                data["arguments"]["jvm"], data, path, options, classpath
            )
        else:
            command.append("-Djava.library.path=" + options["nativesDirectory"])
            command.append("-cp")
            command.append(classpath)
    else:
        command.append("-Djava.library.path=" + options["nativesDirectory"])
        command.append("-cp")
        command.append(classpath)

    # The argument for the logger file
    if options.get("enableLoggingConfig", False):
        if "logging" in data:
            if len(data["logging"]) != 0:
                logger_file = os.path.join(
                    path,
                    "assets",
                    "log_configs",
                    data["logging"]["client"]["file"]["id"],
                )
                command.append(
                    data["logging"]["client"]["argument"].replace(
                        "${path}", logger_file
                    )
                )

    command.append(data["mainClass"])

    if "minecraftArguments" in data:
        # For older versions
        command = command + await get_arguments_string(data, path, options, classpath)
    else:
        command = command + await get_arguments(
            data["arguments"]["game"], data, path, options, classpath
        )

    if "server" in options:
        command.append("--server")
        command.append(options["server"])
        if "port" in options:
            command.append("--port")
            command.append(options["port"])

    if options.get("disableMultiplayer", False):
        command.append("--disableMultiplayer")

    if options.get("disableChat", False):
        command.append("--disableChat")

    return command



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
from typing import TypedDict, Literal


class MrpackFileHashes(TypedDict):
    sha1: str
    sha256: str


class MrpackFileEnv(TypedDict):
    client: Literal["required", "optional", "unsupported"]
    server: Literal["required", "optional", "unsupported"]


class MrpackFile(TypedDict, total=False):
    path: str
    hashes: MrpackFileHashes
    env: MrpackFileEnv
    downloads: list[str]
    fileSize: int


MrpackDependencies = TypedDict(
    "MrpackDependencies",
    {"minecraft": str, "forge": str, "fabric-loader": str, "quilt-loader": str},
    total=False,
)


class MrpackIndex(TypedDict, total=False):
    formatVersion: int
    game: str
    versionId: str
    name: str
    summary: str
    files: list[MrpackFile]
    dependencies: MrpackDependencies



# This file is part of minecraft-launcher-lib (https://codeberg.org/JakobDev/minecraft-launcher-lib)
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 JakobDev <jakobdev@gmx.de> and contributors
# SPDX-License-Identifier: BSD-2-Clause
"""
news includes functions to retrieve news about Minecraft using the official API from Mojang

.. warning::
    The format of the data returned by this API may change at any time
"""
# 標準庫導入
import datetime

# 第三方庫導入
import aiohttp

# 本地導入
from ._types import MinecraftNews, JavaPatchNotes
from ._helper import get_user_agent


async def get_minecraft_news() -> MinecraftNews:
    "Returns general news about Minecraft"
    async with aiohttp.ClientSession() as session:
        async with session.get(
            "https://launchercontent.mojang.com/news.json",
            headers={"user-agent": get_user_agent()},
        ) as response:
            news = await response.json()

    for entry in news["entries"]:
        entry["date"] = datetime.date.fromisoformat(entry["date"])

    return news


async def get_java_patch_notes() -> JavaPatchNotes:
    "Returns the patch notes for Minecraft Java Edition"
    async with aiohttp.ClientSession() as session:
        async with session.get(
            "https://launchercontent.mojang.com/javaPatchNotes.json",
            headers={"user-agent": get_user_agent()},
        ) as response:
            return await response.json()
